import numpy as np
from scipy.fft import fft, ifft, rfft, irfft

# helper function for efficiently taking fft of real or complex fields
def my_fft(u, complex=False):

    if complex:

        out = fft(u)

    else:

        out = rfft(u)

    return out

def my_ifft(V, complex=False):

    if complex:

        out = ifft(V)

    else:

        out = irfft(V)

    return out

# helper function for integration (of real part of fnc) in space. Uses FFT to accurately integrate over spatial domain:
# accuracy vastly beats trapezoidal rule.
# u = array storing node values of field to be integrated (last dimension of the array is spatial)
# length = length of domain
# N = number of samples of u we take (= number of grid pts)
# since this is a postprocessing func it doesn't need to be optimized for real inputs with rfft.
def integrate(u, length, N):
    return (length/N) * np.real(fft(u, axis=-1)[..., 0])



import numpy as np

import joe_lab.joe as joe
import joe_lab.models as joe_models


# get stgrid
length, T, N, dt = 100., 20., 2 ** 10, 1e-3
stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}


# get model
my_model = joe_models.builtin_model('gardner', nonlinear=True)

# get initial state

def gardner_soliton(x, c=1., p=1.):
    return c / (-1. + p * np.sqrt(1. + c) * np.cosh(np.sqrt(c) * x))


def my_initial_state(x):
    c = 2.

    p = -1.

    out = gardner_soliton(x, c=c, p=p)
    return out


my_initial_state = joe.initial_state('gardner_soliton', my_initial_state)

# now define the simulation object
my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=200)

# run it
method_kw = 'etdrk4'
my_sim.load_or_run(method_kw=method_kw, print_runtime=True, save=True)

# produce plots and movies
my_sim.hov_plot(cmap='cmo.haline', fieldname='u', show_figure=True, save_figure=False, usetex=True)
#my_sim.save_movie(dpi=200, fps=200, usetex=False, fieldcolor='xkcd:cerulean', fieldname='u')

# report error in first and second moments, which should both be zero on paper
my_sim.get_fm()
my_sim.get_sm()
fm_error = np.amax(my_sim.fm_error)
print('Maximum error in first moment = %.2E' % fm_error)
sm_error = np.amax(my_sim.sm_error)
print('Maximum error in second moment = %.2E' % sm_error)

#nmin, nmax = 8, 14
#Ns = np.array([2 ** 9, 2 ** 10])
#dts = np.flip(np.logspace(-nmax, -nmin, num=nmax - nmin + 1, base=2.))

#joe.do_refinement_study(my_model, my_initial_state, length, T, Ns, dts, bc='periodic', method_kw=method_kw,
#                    show_figure=True, save_figure=True, usetex=True, fit_min=3, fit_max=9)



# Demo: solving the KdV equation with a sponge layer
# (after Liu and Trogdon 2023: https://www.sciencedirect.com/science/article/pii/S0168927423001526 )

import numpy as np

import joe_lab.joe as joe
import joe_lab.models as joe_models
import joe_lab.initial_states as joe_initial_states

length, T, N, dt = 400., 150., 2**10, 1e-2

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
my_model = joe_models.builtin_model('kdv', nonlinear=True)
my_initial_state = joe_initial_states.builtin_initial_state('gaussian_even_alt')  # also try 'kdv_soliton', 'kdv_multisoliton'

# we need a sponge layer here, which requires parameter tuning to get right.
# in joe, the sponge layer params are contained in a dict called "sponge_params"
# that you pass into the init of your simulation.

l_endpt = -length * 0.5 + 0.5 * length * 0.1
r_endpt = l_endpt + 0.01 * length
width = (2 ** -6) * length / 100.
sponge_params = {'l_endpt': l_endpt, 'r_endpt': r_endpt,
                 'width': width, 'expdamp_freq': 1000,
                 'damping_amplitude': 10.,
                 'splitting_method_kw': 'naive',
                 'spongeless_frac': .5}  # this is the fraction of the middle of the spatial domain to keep in the plots

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='sponge_layer', sponge_params=sponge_params, ndump=20)

#my_sim.plot_initial_condition(color='xkcd:cerulean', usetex=True, show_figure=True, save_figure=True)

my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=False)

# produce plots and movies
my_sim.hov_plot(cmap='cmo.haline', fieldname='u', show_figure=True, save_figure=True, usetex=True)
#my_sim.save_movie(dpi=200, fps=100, usetex=False, fieldcolor='xkcd:cerulean', fieldname='u')
#my_sim.save_combomovie(dpi=200, fps=100, usetex=False, fieldcolor='xkcd:cerulean', speccolor='xkcd:dark magenta', fieldname='u')

# perform refinement study if desired
#nmin, nmax = 2, 14
#Ns = np.array([2**10])
#dts = np.flip(np.logspace(-nmax, -nmin, num=nmax-nmin+1, base=2.))
#joe.do_refinement_study(my_model, my_initial_state, length, T, Ns, dts, bc='sponge_layer',
#                        sponge_params=sponge_params, method_kw='etdrk4', show_figure=True,
#                        save_figure=True, usetex=True, fit_min=2, fit_max=6)



# Code for accuracy assessment, testing how well joe can track a KdV soliton on a periodic grid
# for a long time.

import numpy as np

import joe_lab.joe as joe
import joe_lab.models as joe_models
import joe_lab.initial_states as joe_initial_states

length, T = 100., 180.
my_model = joe_models.builtin_model('kdv', nonlinear=True)
my_initial_state = joe_initial_states.builtin_initial_state('kdv_soliton')
method_kw = 'etdrk4'

nmin, nmax = 2, 14
Ns = np.array([2**7, 2**8])
dts = np.flip(np.logspace(-nmax, -nmin, num=nmax-nmin+1, base=2.))

joe.do_refinement_study(my_model, my_initial_state, length, T, Ns, dts, bc='periodic', method_kw=method_kw,
                    show_figure=True, save_figure=True, usetex=True, fit_min=3, fit_max=9)



import numpy as np
from scipy.fft import rfft, irfft, rfftfreq

import joe_lab.joe as joe
import joe_lab.models as joe_models
import joe_lab.initial_states as joe_initial_states
from joe_lab.visualization import nice_plot
from joe_lab.utils import integrate

# first prescribe all the simulation parameters etc.
length, T, N, dt = 240.,100., 2**9, 1e-2

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
my_model = joe_models.builtin_model('phi4pert', nonlinear=True)
my_initial_state = joe_initial_states.builtin_initial_state('gaussian_odd')

# obtain the relevant simulation, either by loading it up again or by running it.
my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=20)
my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=True)

# produce plots and movies
my_sim.hov_plot(cmap='cmo.haline', fieldname='u', show_figure=True, save_figure=True, usetex=True)
#my_sim.save_movie(dpi=200, fps=200, usetex=False, fieldcolor='xkcd:cerulean', fieldname='u')

# Now we do all the post-processing energy analysis, and an accuracy test

# Define a function that computes the energy of our state [u,ut] at time t
x = my_sim.x
Udata = my_sim.Udata

# need these for energy, and they're built-in to make the nonlinear term for phi^4 perturbations
from joe_lab.models import K0, V0

def energy(u, ut):
    kin = ut ** 2

    # get wavenumbers for the grid of S^1 with N samples
    k = 2. * np.pi * N * rfftfreq(N) / length

    spring = irfft(1j * k * rfft(u)) ** 2

    potential = (2. + V0(x)) * (u ** 2) + 1.*(2. * K0(x) * u ** 3 + 0.5 * u ** 4)

    e_dens = kin + spring + potential

    out = 0.5*integrate(e_dens, length, N)

    return out

# get the energies associated to each time
times = np.linspace(0., T, num=1 + int(T / (dt*my_sim.ndump)), endpoint=True)
E = energy(Udata[0, :, :],  Udata[1, :, :])

# draw the figure
dpi = 400
picname = 'phi4pert_energy_test_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '_ICkw=' + my_sim.initial_state_kw  + '.png'
nice_plot(times, (E-E[0])/E[0], r'$t$', r'Relative Error in Energy', dpi=dpi, show_figure=True,
          save_figure=True, picname=picname, linestyle='solid', color='xkcd:blueberry', usetex=True)

# perform Richardson accuracy test
nmin, nmax = 4, 9
Ns = np.array([2**7, 2**8, 2**9, 2**10])
dts = np.flip(np.logspace(-nmax, -nmin, num=nmax-nmin+1, base=2.))

do_refinement_study(my_model, my_initial_state, length, T, Ns, dts, bc='periodic', method_kw='etdrk4',
                    show_figure=False, save_figure=True, usetex=True, dpi=dpi, fit_min=0, fit_max=4)



import numpy as np

from .joe import initial_state


def kdv_soliton(x, c=1.):
    if c <= 0:
        raise ValueError('KdV soliton speed must be strictly positive.')
    return 0.5*c*(np.cosh(0.5 * np.sqrt(c) * x) ** -2)


def gardner_soliton(x, c=1., p=1.):
    if c <= 0:
        raise ValueError('Focusing Gardner soliton speed must be strictly positive.')
    out = np.zeros_like(x, dtype=float)
    xmax = 180
    out[abs(x) > xmax] = 0.
    out[abs(x) <= xmax] = c / (-1. + p * np.sqrt(1. + c) * np.cosh(np.sqrt(c) * x[abs(x) <= xmax]))
    return out


def bbm_solitary_wave(x, c=1.):
    if c <= 0 and c >= -1:
        raise ValueError('BBM soliton speed must be strictly positive, or strictly less than -1.')
    return 0.5*c*(np.cosh(0.5 * np.sqrt(c/(1.+c)) * x) ** -2)

def sinegordon_soliton(x, c=0., p=1.):
    if np.abs(c) >= 1:
        raise ValueError('Sine-Gordon soliton speed must be in the interval (-1,1).')
    out = np.zeros_like(x, dtype=float)
    arg = p*x/np.sqrt(1.-c**2)
    xmax = 1.8e2
    out[arg > xmax] = 2.*np.pi
    out[arg < -xmax] = 0.
    out[np.abs(arg)<= xmax] = 4.*np.arctan(np.exp(arg[np.abs(arg)<= xmax]))
    return out

def sinegordon_soliton_speed(x, c=0., p=1.): # speed is at t=0
    if np.abs(c) >= 1:
        raise ValueError('Sine-Gordon soliton speed must be in the interval (-1,1).')
    out = np.zeros_like(x, dtype=float)
    gamma = 1./np.sqrt(1.-c**2)
    arg = p*x*gamma
    xmax = 1.8e2
    out[arg > xmax] = 0.
    out[arg < -xmax] = 0.
    aa = arg[np.abs(arg)<= xmax]
    out[np.abs(arg)<= xmax] = -(c*4.*p*gamma)*(np.exp(aa)/(1.+np.exp(2.*aa)))
    return out


def initial_state_func(x, initial_state_kw):
    amp = 0.1
    x0 = 0.
    k0 = 1.
    width = 1.

    if initial_state_kw == 'sine':
        k = 1.5
        out = 1.0*np.sin(k*x)

    elif initial_state_kw == 'gaussian_even':
        out = 6.*np.exp(-x**2)

    elif initial_state_kw == 'gaussian_even_alt':
        out = 1.3*np.exp(-x**2)

    elif initial_state_kw == 'kdv_soliton':
        c = 2.
        out = kdv_soliton(x, c=c)

    elif initial_state_kw == 'kdv_multisoliton':
        c0 = 3.2
        c1 = 2.5
        c2 = 1.
        out = kdv_soliton(x+80., c=c0) + kdv_soliton(x+50., c=c1) + kdv_soliton(x+10, c=c2)

    elif initial_state_kw == 'gardner_soliton':
        c = 3.
        p = 1.
        out = gardner_soliton(x, c=c, p=p)

    elif initial_state_kw == 'bbm_solitary_wave':
        c = 2.
        out = bbm_solitary_wave(x,c=c)

    elif initial_state_kw == 'bbm_multisolitary':
        c0 = 2.5
        c1 = 2.
        out = bbm_solitary_wave(x+80., c=c0) + bbm_solitary_wave(x+50., c=c1)

    elif initial_state_kw == 'gaussian_odd':
        out = amp * (np.sin(k0 * x)) * np.exp(-width * (x - x0) ** 2)

    elif initial_state_kw == 'gaussian_no_parity':
        out = amp * (0.7 * np.sin(k0 * x) + 0.3 * np.cos(x)) * np.exp(-width * (x - x0) ** 2)

    elif initial_state_kw == 'translational_mode':
        out = np.cosh(x / np.sqrt(2)) ** -2

    elif initial_state_kw == 'internal_mode':
        out = amp*np.sinh(x / np.sqrt(2)) * (np.cosh(x / np.sqrt(2))) ** -2

    elif initial_state_kw == 'tritone':
        a = 1.2*np.sqrt(2.)  # this value gives the Getmanov tri-tone!
        out = a*np.sinh(x / np.sqrt(2)) * (np.cosh(x / np.sqrt(2))) ** -2

    elif initial_state_kw == 'trivial':
        pass

    elif initial_state_kw == '0_energy':
        out = -1. + 3. * np.tanh(x / np.sqrt(2)) ** 2

    elif initial_state_kw == 'ks_chaos':
        out = np.cos((x+16.*np.pi)/16.) * (1. + np.sin((x+16.*np.pi) / 16.))

    elif initial_state_kw == 'bbm_weird_wavepacket':
        out = (0.1*np.cos(20.*x) + np.cos(0.2*x))*np.exp(-x**2)

    elif initial_state_kw == 'sinegordon_soliton_interaction':
        out = np.zeros((2,np.shape(x)[0]), dtype=float)
        out[0,:] = sinegordon_soliton(x+20, c=0.9, p=1) - sinegordon_soliton(x-20, c=-0.9, p=1)
        out[1,:] = sinegordon_soliton_speed(x+20, c=0.9, p=1) - sinegordon_soliton_speed(x-20, c=-0.9, p=1)

    elif initial_state_kw == 'sinegordon_soliton_interaction_alt':
        out = np.zeros((2,np.shape(x)[0]), dtype=float)
        out[0,:] = sinegordon_soliton(x+20, c=0.9, p=1) - sinegordon_soliton(x-20, c=0., p=1)
        out[1,:] = sinegordon_soliton_speed(x+20, c=0.9, p=1)

    else:
        raise NameError("User-defined initial state keyword string not among built-in options.")

    return out

def builtin_initial_state(initial_state_kw):

    return initial_state(initial_state_kw, lambda x : initial_state_func(x, initial_state_kw))
    



import numpy as np

import joe_lab.joe as joe
import joe_lab.models as joe_models
import joe_lab.initial_states as joe_initial_states

length, T, N, dt = 100., 50., 2**10, 1e-2

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
my_model = joe_models.builtin_model('bbm', nonlinear=True)
my_initial_state = joe_initial_states.builtin_initial_state('bbm_solitary_wave')

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=20)

#my_sim.plot_initial_condition(color='xkcd:cerulean', usetex=True, show_figure=True, save_figure=True)

my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=False)

# produce plots and movies
my_sim.hov_plot(cmap='cmo.haline', fieldname='u', show_figure=True, save_figure=True, usetex=True)
#my_sim.save_movie(dpi=200, fps=100, usetex=False, fieldcolor='xkcd:cerulean', fieldname='u')
#my_sim.save_combomovie(dpi=200, fps=100, usetex=False, fieldcolor='xkcd:cerulean', speccolor='xkcd:dark magenta', fieldname='u')

# report error in first moment
my_sim.get_fm()
error = np.amax(my_sim.fm_error)
print('Maximum error in first moment = %.2E' % error)



import numpy as np

from scipy.fft import fft, ifft, fftfreq, rfft, irfft, rfftfreq

from .joe import model

# Aux functions needed for special cases...

# obtain the kink
def K0(x):
    out = np.tanh(x / np.sqrt(2))

    return out


# obtain the potential associated to the kink
# Note: the +2 in the potential gets put into linear part of evolution eq.
def V0(x):
    out = -3. * np.cosh(x / np.sqrt(2)) ** -2

    return out


# Here begins the actual core of the material

# if model is first order in time, below just gives linear, const. coeff. part. in Fourier space
# if model is second order in time, instead obtain the spatial operator for the first order system as a block matrix
def get_symbol(k, model_kw):

    if model_kw == 'phi4pert':
        A = -(k ** 2 + 2. * np.ones_like(k))

    elif model_kw =='sinegordon':

        A = -k**2

    elif model_kw == 'bbm' or model_kw == 'gardner-bbm':
        A = 1j * (k ** 3) / (1. + k ** 2)

    elif model_kw == 'ks':
        A = k ** 2 - k ** 4

    elif model_kw == 'kdv' or model_kw == 'gardner':
        A = 1j * k ** 3

    else:

        raise NameError("Invalid model keyword string.")

    return A


def fourier_forcing(V, k, x, model_kw, nonlinear=True):
    # Fourier transform of forcing term, acting on pair fncs V=(v_1, v_2)^T (concatenation)
    # on Fourier space. V has size 2N if complex, or size N+2 if real

    if model_kw == 'phi4pert':

        if int(V.size-2) == x.size:

            pass

        else:

            raise TypeError("The array V must be 2+(size of the array x) if our soln is real, "
                            "or 2*(size of the array x) if our soln is complex."
                            " Size of V = ", int(V.size), "size of x = ", x.size)

        N = int(V.size-2)

        V = np.reshape(V, (N+2,))

        NN = int(0.5*N)+1

        u = irfft(V[0:NN])  # only ifft first N entries of V because of storage conventions

        spatial_forcing = -1. * V0(x) * u - float(nonlinear) * (3. * K0(x) * u ** 2 + u ** 3)

        out = 1j * np.zeros(N+2, dtype=float)
        out[NN:] = rfft(spatial_forcing)

    elif model_kw == 'sinegordon':

        if int(V.size - 2) == x.size:

            pass

        else:

            raise TypeError("The array V must be 2+(size of the array x) if our soln is real, "
                            "or 2*(size of the array x) if our soln is complex."
                            " Size of V = ", int(V.size), "size of x = ", x.size)

        N = int(V.size - 2)

        V = np.reshape(V, (N + 2,))

        NN = int(0.5 * N) + 1

        u = irfft(V[0:NN])  # only ifft first N entries of V because of storage conventions

        spatial_forcing = -float(nonlinear)*np.sin(u)

        out = 1j * np.zeros(N + 2, dtype=float)
        out[NN:] = rfft(spatial_forcing)

    elif model_kw == 'bbm':

        p = 1.

        out = -6. * float(nonlinear) * (1. / (p + 1.)) * (1j * k / (1. + k ** 2) )*  rfft(irfft(V) ** (p + 1))

    elif model_kw == 'ks':

        p = 1.

        out = -float(nonlinear) * (1. / (p + 1.)) * 1j * k * rfft(irfft(V) ** (p + 1))

    elif model_kw == 'gardner':

        out = 6. * float(nonlinear) * (1j * k) * (
                    0.5 * rfft(irfft(V) ** 2) - (1. / 3.) * rfft(irfft(V) ** 3))

    elif model_kw == 'gardner-bbm':

        out =  6. * float(nonlinear) * ((1j * k)/(1. + k**2)) *  (
                    0.5 * rfft(irfft(V) ** 2) - (1. / 3.) * rfft(irfft(V) ** 3))

    elif model_kw == 'kdv':

        p = 1.

        out = -6. * float(nonlinear) * (1. / (p + 1.)) * 1j * k * rfft(irfft(V) ** (p + 1))

    else:

        raise NameError("Invalid model keyword string.")

    return out


def builtin_model(model_kw, nonlinear=True):

    def my_symbol(k):
        return get_symbol(k, model_kw)

    if model_kw == 'phi4pert' or model_kw == 'sinegordon':

        t_ord = 2

        def my_fourier_forcing(V, k, x, nonlinear):
            return fourier_forcing(V, k, x, model_kw, nonlinear)

    else:

        t_ord = 1

        def my_fourier_forcing(V, k, x, nonlinear):
            return fourier_forcing(V, k, x, model_kw, nonlinear)

    return model(model_kw, t_ord, my_symbol, my_fourier_forcing, nonlinear=nonlinear)



import pickle
import os

import numpy as np
from scipy.fft import fftfreq, rfftfreq
from scipy import sparse
from scipy.sparse import linalg

from .utils import my_fft, my_ifft
from .sponge_layer import damping_coeff_lt, rayleigh_damping


# The intention here is to make the code independent of the particular
# PDE we're considering insofar as is possible.

# First, a function for computing all the Greeks ("weights" for exponential quadrature).
# We do this by Pythonizing the code from Kassam and Trefethen 2005 (do Cauchy integrals).

def get_greeks_first_order(N, dt, z):
    M = 2 ** 6

    theta = np.linspace(0., 2. * np.pi, num=M, endpoint=False)

    rad = 1.

    z0 = dt * np.tile(z, (M, 1)) + rad * np.tile(np.exp(1j * theta), (N, 1)).T

    Q = dt * np.mean((np.exp(0.5 * z0) - 1.) / z0, 0)  # note how we take mean over a certain axis

    f1 = dt * np.mean((-4. - z0 + np.exp(z0) * (4. - 3. * z0 + z0 ** 2)) / (z0 ** 3), 0)  # again note axis
    # argument of np.mean

    f2 = dt * np.mean((2. + z0 + np.exp(z0) * (-2. + z0)) / (z0 ** 3), 0)

    f3 = dt * np.mean((-4. - 3. * z0 - z0 ** 2 + np.exp(z0) * (4. - z0)) / (z0 ** 3), 0)

    out = [Q, f1, f2, f3]

    return out


def get_greeks_second_order(length, N, dt, A):
    M = 2 ** 6
    theta = np.linspace(0, 2. * np.pi, num=M, endpoint=False)

    # radius of contour = largest eigenvalue of linear part with a bit of wiggle room
    # a bit of analysis shows the largest eigval is sqrt(A(k_max)) where k_max
    # is the largest positive frequency allowed.
    rad = 1.2 * dt * np.sqrt(np.amax(np.abs(A)))
    r = rad * np.exp(1j * theta)

    id_matrix = sparse.eye(2 * N, dtype=float)

    Q = 1j * np.zeros([2 * N, 2 * N], dtype=float)
    f1 = 1j * np.zeros([2 * N, 2 * N], dtype=float)
    f2 = 1j * np.zeros([2 * N, 2 * N], dtype=float)
    f3 = 1j * np.zeros([2 * N, 2 * N], dtype=float)

    for j in np.arange(0, M):
        z = r[j]

        B = id_matrix.multiply(z) - A.multiply(dt)

        B = sparse.csc_matrix(B)

        zIA = sparse.linalg.inv(B)

        Q += dt * zIA * (np.exp(0.5 * z) - 1.)
        f1 += dt * zIA * ((-4. - z + np.exp(z) * (4. - 3. * z + z ** 2)) / (z ** 2))
        f2 += dt * zIA * ((2. + z + np.exp(z) * (-2. + z)) / (z ** 2))
        f3 += dt * zIA * ((-4. - 3. * z - z ** 2 + np.exp(z) * (4. - z)) / (z ** 2))

    # TODO: I think the "real" below may eventually cause issues, but then again
    #  I don't think there are any interesting complex-valued second order eqns...maybe come back to this!
    #  Perhaps it's worth trying for, say, the zero-magnetic field abelian Higgs equations?
    Q = sparse.csc_matrix(np.real(Q / M))
    f1 = sparse.csc_matrix(np.real(f1 / M))
    f2 = sparse.csc_matrix(np.real(f2 / M))
    f3 = sparse.csc_matrix(np.real(f3 / M))

    out = [Q, f1, f2, f3]

    return out


def get_Q1(N, dt, z):
    M = 2 ** 6

    theta = np.linspace(0, 2. * np.pi, num=M, endpoint=False)

    rad = 1.  # radius of contour about dt*z about which we integrate

    z0 = dt * np.tile(z, (M, 1)) + rad * np.tile(np.exp(1j * theta), (N, 1)).T

    out = dt * np.mean((np.exp(z0) - 1.) / z0, 0)  # note how we take mean over a certain axis

    return out


def get_R2(N, dt, z):
    # """
    M = 2 ** 5

    theta = np.linspace(0, 2. * np.pi, num=M, endpoint=False)

    rad = 1.  # radius of contour about dt*z about which we integrate

    z0 = dt * np.tile(z, (M, 1)) + rad * np.tile(np.exp(1j * theta), (N, 1)).T

    out = dt * np.mean((np.exp(z0) - 1. - z0) / (z0 ** 2), 0) # note how we take mean over a certain axis
    # """

    return out


def do_etdrk1_step(V, propagator, forcing, Q1):
    # remark on notation: Q1 = dt*phi1(dt*A)

    out = propagator * V + Q1 * forcing(V)

    return out


def do_etdrk2_step(V, propagator, forcing, Q1, R2):
    a = do_etdrk1_step(V, propagator, forcing, Q1)

    out = a + R2 * (forcing(a) - forcing(V))

    return out


# code for a single ETDRK4 step, for a first-order-in-time PDE
def do_etdrk4_step(V, propagator, propagator2, forcing, greeks):
    Q = greeks['Q']
    f1 = greeks['f1']
    f2 = greeks['f2']
    f3 = greeks['f3']

    fV = forcing(V)

    Vhalf = propagator2 * V

    a = Vhalf + Q * fV

    fa = forcing(a)

    b = Vhalf + Q * fa

    fb = forcing(b)

    c = propagator2 * a + Q * (2. * fb - fV)

    fc = forcing(c)

    # now assemble the guess at the new step
    out = propagator * V + f1 * fV + 2. * f2 * (fa + fb) + f3 * fc

    return out


# code for a single ETDRK4 step, if the eqn is second order in time (basically the above code, but sparsified)
def do_etdrk4_step_second_order(V, propagator, propagator2, forcing, greeks):
    Q = greeks['Q']
    f1 = greeks['f1']
    f2 = greeks['f2']
    f3 = greeks['f3']

    N = int(0.5 * np.size(V))

    fV = forcing(V)

    Vhalf = propagator2 @ V  # note: @ takes advantage of sparsity.

    a = Vhalf + np.asarray(Q @ fV)

    a = np.reshape(a, (2 * N,))

    fa = forcing(a)

    b = Vhalf + np.asarray(Q @ fa)

    b = np.reshape(b, (2 * N,))

    fb = forcing(b)

    c = np.asarray(propagator2 @ a + Q @ (2. * fb - fV))

    c = np.reshape(c, (2 * N,))

    fc = forcing(c)

    # now assemble the guess at the new step. This is the temporal bottleneck of the time step (probably like 70% of step time)
    out = np.asarray(propagator @ V + f1 @ fV + 2. * f2 @ (fa + fb) + f3 @ fc)

    out = np.reshape(out, (2 * N,))

    return out


# code for a single integrating factor Runge-Kutta fourth-order step
def do_ifrk4_step(V, propagator, propagator2, forcing, dt):
    a = dt * forcing(V)

    b = dt * forcing(propagator2 * (V + 0.5 * a))

    c = dt * forcing(propagator2 * V + 0.5 * b)

    d = dt * forcing(propagator * V + propagator2 * c)

    out = propagator * V + (1. / 6.) * (propagator * a + 2. * propagator2 * (b + c) + d)

    return out


def assemble_damping_mat(N, length, x, dt, sponge_params, complex = False):
    # By "damping mat", we mean the matrix to be inverted at each time step in the damping stage.
    # Currently only backward Euler inversion is implemented.
    # TODO: try out Crank-Nicolson as well, perform cost v. accuracy analysis?

    if complex:

        k = 2 * np.pi * N * fftfreq(N) / length

    else:

        k = 2 * np.pi * N * rfftfreq(N) / length

    # Deal w/ the damping mat as a scipy.sparse LinearOperator to avoid matrix mults!
    # This is an issue here bcz dealing with the matrix of the Fourier transform is a pain

    def mv(v):
        # NOTE: v is given ON THE FOURIER SIDE!!!!

        damping_coeff = damping_coeff_lt(x, sponge_params) #+ damping_coeff_lt(-x, sponge_params)

        damping_coeff *= sponge_params['damping_amplitude']

        mv_out = v - dt * (-1j * k) * my_fft(damping_coeff * my_ifft((-1j * k) * v, complex=complex), complex=complex)

        return mv_out

    NN = k.size

    out = linalg.LinearOperator(shape=(NN, NN), matvec=mv)

    # TODO: apparently LinearOperators can't be pickled, but sparse matrices can? See if there is some other decent
    # way of saving LinearOperators

    return out


def do_diffusion_step(q, B):
    B_LHS = B

    RHS = q

    out, info = linalg.cg(B_LHS, RHS)  # have to have "info" here otherwise the code throws a fit

    return out


class timestepper:
    def __init__(self, method_kw, sim, scale=1., complex=False):
        self.method_kw = method_kw
        self.sim = sim
        self.scale = scale
        self.t_ord = sim.model.t_ord
        self.complex = complex
        self.aux = None
        dt_new = scale*sim.dt
        self.auxfilename = 'timestepper_auxfile_method_kw=' + self.method_kw + '_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (
            sim.length, sim.T, sim.N,
            dt_new) + '_modelkw=' + sim.model_kw + '.pkl'  # need to save as pkl since we store aux as a dict

    def get_aux(self):

        sim = self.sim

        t_ord = self.t_ord

        length = sim.length

        N = sim.N

        dt = self.scale*sim.dt

        if self.complex:

            k = 2. * np.pi * N * fftfreq(N) / length

            NN = N

        else:

            k = 2. * np.pi * N * rfftfreq(N) / length

            NN = int(0.5*N +1) # only keep the real frequencies

        A = sim.model.get_symbol(k)

        if t_ord == 1:

            propagator = np.exp(dt * A)

        elif t_ord == 2:

            A = sparse.diags([A, np.ones(NN, dtype=float)], [-NN, NN], shape=[2 * NN, 2 * NN]).tocsc()

            propagator = linalg.expm(A.multiply(dt))

        if self.method_kw == 'etdrk1':
            Q1 = get_Q1(NN, dt, A)

            aux = dict([('Q1', Q1), ('propagator', propagator)])

        if self.method_kw == 'etdrk4':

            if t_ord == 1:

                [Q, f1, f2, f3] = get_greeks_first_order(NN, dt, A)

                propagator2 = np.exp(0.5 * dt * A)

            elif t_ord == 2:

                [Q, f1, f2, f3] = get_greeks_second_order(length, NN, dt, A)

                propagator2 = linalg.expm(A.multiply(0.5*dt))

            aux = dict([('Q', Q), ('f1', f1), ('f2', f2), ('f3', f3), ('propagator', propagator),
                        ('propagator2', propagator2)])

        if self.method_kw == 'ifrk4':
            propagator2 = np.exp(0.5 * dt * A)

            aux = dict([('propagator', propagator), ('propagator2', propagator2)])

        self.aux = aux

    def save_aux(self):

        # add the folder "joe_timestepper_aux" to our path... more on this below
        my_path = os.path.join("joe_timestepper_aux")

        # first, if the folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        with open('joe_timestepper_aux/'+self.auxfilename, 'wb') as outp:
            pickle.dump(self.aux, outp, pickle.HIGHEST_PROTOCOL)

    def load_aux(self):

        with open('joe_timestepper_aux/'+self.auxfilename, 'rb') as inp:
            self.aux = pickle.load(inp)

    def do_time_step(self, V, forcing):

        t_ord = self.t_ord

        aux = self.aux

        propagator = aux['propagator']

        if self.method_kw == 'etdrk1':
            Q1 = aux['Q1']

            out = do_etdrk1_step(V, propagator, forcing, Q1)

        if self.method_kw == 'etdrk4':
            Q = aux['Q']

            f1 = aux['f1']

            f2 = aux['f2']

            f3 = aux['f3']

            greeks = dict([('Q', Q), ('f1', f1), ('f2', f2), ('f3', f3)])

            propagator = aux['propagator']

            propagator2 = aux['propagator2']

            if t_ord == 1:

                out = do_etdrk4_step(V, propagator, propagator2, forcing, greeks)

            if t_ord == 2:

                out = do_etdrk4_step_second_order(V, propagator, propagator2, forcing, greeks)

        if self.method_kw == 'ifrk4':
            propagator = aux['propagator']

            propagator2 = aux['propagator2']

            out = do_ifrk4_step(V, propagator, propagator2, forcing, self.scale*self.sim.dt)

        return out


def do_time_stepping(sim, method_kw='etdrk4'):
    length = sim.length

    T = sim.T

    N = sim.N

    dt = sim.dt

    model = sim.model

    t_ord = model.t_ord

    if t_ord == 2 and method_kw != 'etdrk4':
        raise ValueError('Only ETDRK4 time-stepping is currently supported for second-order equations.')

    complex = sim.complex

    initial_state = sim.initial_state

    nonlinear = sim.nonlinear

    sponge_layer = sim.sponge_layer

    if sponge_layer:

        sponge_params = sim.sponge_params

        if t_ord == 1:

            splitting_method_kw = sponge_params['splitting_method_kw']

        else:

            splitting_method_kw = 'na' # Rayleigh damping for second order means no splitting required

    else:

        splitting_method_kw = 'na'

    # account for the difference between real and complex during storage

    if complex:

        NN = N

    else:

        NN = int(0.5*N)+1

    ndump = sim.ndump

    nsteps = int(T / dt)

    x = sim.x  # the endpoint = False flag is critical!

    if complex:

        k = 2. * np.pi * N * fftfreq(N) / length

    else:

        k = 2. * np.pi * N * rfftfreq(N) / length

    # determine the time-step scale factor "a" for splitting
    if splitting_method_kw == 'strang':

        scale = 0.5

    else:

        scale = 1.

    my_timestepper = timestepper(method_kw, sim, scale=scale, complex=complex)

    # preprocessing stage: assemble the aux quantities needed for time-stepping, and the forcing function

    # create forcing term

    def forcing(V):

        out = model.get_fourier_forcing(V, k, x, nonlinear)

        # if we're second-order in time and using a sponge layer, damping can be realized simply
        # by modifying the forcing term ie. damping can be dealt with explicitly!
        if t_ord == 2 and sponge_layer:

            out += rayleigh_damping(V, x, length, sponge_params, complex=complex)

        return out

    # obtain the aux quantities. Thanks to all the hard work we did when defining the timestepper class, the code here
    # is brief and (IMO) elegant.

    # first check if we've already computed aux on the required space-time grid
    try:
        my_timestepper.load_aux()

    # if the auxfile is not found, compute aux here.
    except FileNotFoundError:

        my_timestepper.get_aux()
        my_timestepper.save_aux()

    # now assemble the stuff needed for damping
    if sponge_layer and t_ord == 1:
        damping_mat = assemble_damping_mat(N, length, x, dt, sponge_params, complex=complex)
    else:
        pass

    # now set up the initial conditions
    Uinit = initial_state

    if t_ord == 2:

        try:

            v1 = my_fft(Uinit[0, :], complex=complex)
            v2 = my_fft(Uinit[1, :], complex=complex)

        except: # if no initial speed is provided in second order case, default to assuming it's zero.

            v1 = my_fft(Uinit, complex=complex)
            v2 = 1j*np.zeros_like(v1, dtype=float)

        V = np.concatenate((v1, v2))

        # make data storage array

        if complex:

            Udata = 1j*np.zeros([2, 1 + int(nsteps / ndump), N], dtype=float)

        else:

            Udata = np.zeros([2, 1 + int(nsteps / ndump), N], dtype=float)

        try:

            Udata[0, 0, :] = Uinit[0, :]
            Udata[1, 0, :] = Uinit[1, :]

        except: # if no initial speed is provided in second order case, default to assuming it's zero.

            Udata[0, 0, :] = Uinit
            pass

    elif t_ord == 1:

        V = my_fft(Uinit, complex=complex)

        # make data storage array
        if complex:

            Udata = 1j*np.zeros([1 + int(nsteps / ndump), N], dtype=float)

        else:

            Udata = np.zeros([1 + int(nsteps / ndump), N], dtype=float)

        Udata[0, :] = Uinit

    else:

        raise ValueError('t_ord must be equal to 1 or 2!')

    cnt = 0.  # counter

    for n in np.arange(1, nsteps + 1):

        Va = my_timestepper.do_time_step(V, forcing)

        if sponge_layer and t_ord == 1:

            if splitting_method_kw == 'naive':

                Vb = do_diffusion_step(Va, damping_mat)

                V = Vb

            elif splitting_method_kw == 'strang':

                Vb = do_diffusion_step(Va, damping_mat)

                Vc = my_timestepper.do_time_step(Vb, forcing)

                V = Vc

            if cnt % int(sponge_params['expdamp_freq']) == 0:

                U = my_ifft(V, complex=complex)

                U *= 1. - 1. * damping_coeff_lt(-x, sponge_params) - 1. * damping_coeff_lt(x, sponge_params)

                V = my_fft(U, complex=complex)

        else:

            V = Va

        cnt += 1

        # data storage step
        if cnt % ndump == 0:

            if t_ord == 2:

               Udata[0, int(n / ndump), :] = my_ifft(V[0:NN], complex=complex)
               Udata[1, int(n / ndump), :] = my_ifft(V[NN:], complex=complex)

            else:

                Udata[int(n / ndump), :] = my_ifft(V, complex=complex)

        else:

            pass

    return Udata



import pickle
import os

import numpy as np
from numpy.fft import fft
import matplotlib.pyplot as plt

from .time_stepper import do_time_stepping
from .visualization import hov_plot, save_movie, save_combomovie, spinner, plot_refinement_study, nice_plot
from .sponge_layer import clip_spongeless
from .utils import integrate
# a class for models. Note how the init takes in two callables for the symbol and forcing terms: to avoid making the
# weird no-no of having callable attributes, we use a trick from
# https://stackoverflow.com/questions/35321744/python-function-as-class-attribute-becomes-a-bound-method
# and instead make the "callable attributes" !dicts! with callable entries. This dirty trick is all under the hood
# in the model class, so when defining and using a model object the user doesn't need to care.
class model:
    def __init__(self, model_kw, t_ord, symbol, fourier_forcing, nonlinear=True, complex=False):
        self.model_kw = model_kw
        self.t_ord = t_ord  # an integer
        self.symbol = {'symbol': symbol}  # callable
        self.fourier_forcing = {'fourier_forcing': fourier_forcing}  # callable
        self.nonlinear = nonlinear
        self.complex = complex

        # this defines a model PDE with the name 'model_kw' in the Fourier-space form

        # (d_t)^{t_ord} V + symbol*V = fourier_forcing

        # where V is the Fourier state.

    # obtain the actual symbol on a given mesh
    def get_symbol(self, *args):
        return self.symbol['symbol'](*args)

    # obtain the forcing term in Fourier space
    def get_fourier_forcing(self, *args):
        return self.fourier_forcing['fourier_forcing'](*args)


# turn initial states into object as well, following the syntax used to define the model class. This is really there
# instead of kw's only just to allow the user to easy input custom initial states on the fly.
class initial_state:
    def __init__(self, initial_state_kw, initial_state_func):
        self.initial_state_kw = initial_state_kw
        self.initial_state_func = {'initial_state_func': initial_state_func}
        self.initial_state = None

    # obtain the actual initial state fnc
    def get_initial_state(self, *args):
        return self.initial_state_func['initial_state_func'](*args)


# a class for simulations. You init with a "stgrid" (space-time grid) dict, a "model" object to specify the PDE,
# and an "initial_state" object. Then you can call a run simulation function on a simulation object to solve
# the IVP and store/save the soln, or load up the soln from an experiment you've already done.

class simulation:
    def __init__(self, stgrid, model, initial_state, bc, sponge_params=None, ndump=10):
        self.length = stgrid['length']
        self.T = stgrid['T']
        self.N = stgrid['N']
        self.dt = stgrid['dt']
        self.model = model  # a model object
        self.model_kw = model.model_kw
        self.t_ord = model.t_ord  # an integer
        self.initial_state_kw = initial_state.initial_state_kw
        self.nonlinear = model.nonlinear
        self.sponge_params = sponge_params
        self.complex = model.complex

        if bc == 'sponge_layer':
            self.sponge_layer = True
        elif bc == 'periodic':
            self.sponge_layer = False
        else:
            raise ValueError('User-defined BC string not accepted. Valid BC strings: periodic, sponge_layer')

        # the "spongeless fraction" attribute is a bit special for plotting and so deserves to be
        # singled out early on
        try:
            self.sfrac = self.sponge_params['spongeless_frac']
        except TypeError:
            self.sfrac = 1.


        self.ndump = ndump  # hyperparameter describing how often we save our time steps
        self.x = np.linspace(-0.5 * self.length, 0.5 * self.length, self.N, endpoint=False)
        self.initial_state = initial_state.get_initial_state(self.x) # IMPORTANT: self.initial_state is an array, but the
        # actual input initial_state to the simulation class is an initial_state object!

        my_string = '_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (self.length, self.T, self.N,
                                                                       self.dt) + '_modelkw=' + self.model_kw + '_ICkw=' + self.initial_state_kw + '_nonlinear=' + str(
            self.nonlinear) + '_sponge_layer=' + str(self.sponge_layer)

        self.filename = 'simdata' + my_string + '.pkl'
        self.ic_picname = 'ic_plot' + my_string + '.png'
        self.picname = 'hovplot' + my_string + '.png'
        self.realpicname = 'hovplot_real' + my_string + '.png'
        self.imagpicname = 'hovplot_imag' + my_string + '.png'
        self.modpicname = 'hovplot_mod' + my_string + '.png'
        self.moviename = 'movie' + my_string + '.mp4'
        self.realmoviename = 'movie_real' + my_string + '.mp4'
        self.imagmoviename = 'movie_imag' + my_string + '.mp4'
        self.modmoviename = 'movie_mod' + my_string + '.mp4'
        self.combomoviename = 'combomovie' + my_string + '.mp4'
        self.Udata = None  # the Udata will be called later!
        self.fm = None
        self.fm_error = None
        self.sm = None
        self.sm_error = None

    # a function for actually performing the time-stepping on a simulation object. Adds the property Udata
    # to the simulation object (the actual values of our solution throughout the simulation)
    def run_sim(self, method_kw='etdrk4', print_runtime=True):
        import time
        start = time.time()
        Udata = do_time_stepping(self, method_kw)
        end = time.time()

        self.Udata = Udata

        if print_runtime:
            runtime = end - start
            print('Simulation runtime = %.3f' % runtime, 's')

    # save the simulation object to an external .pkl file using the pickle module.
    def save(self):
        self.model = None # OK this line needs some explaining! Basically the simulation object needs to track its
        # model, and not just the model_kw, bcz during time-stepping we need to of course access the symbol and the
        # forcing term. BUT to make the built-in (non-custom) models easily callable from just the model_kw, we need
        # to define model objects that in turn involve nested functions. And these can't be pickled because of course not
        # why would anything work. So, the best way I found to accommodate all of...
        # 1) having the model available for time-stepping
        # 2) being able to have users define custom models
        # 3) being able to have users just call one of my built-in models with a model_kw
        # was to just forget the actual model attribute of our simulation. Since we still keep the model_kw attribute
        # this is not a big deal when it comes to saving/loading: as long as we have all of the data we have from the sim
        # and the file is named properly, there's nothing to worry about. Said differently, we keep the full model attribute
        # of a simulation object around only as long as we need it.

        my_path = os.path.join("joe_sim_archive")

        # if the archive folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        with open('joe_sim_archive/'+self.filename, 'wb') as outp:
            pickle.dump(self, outp, pickle.HIGHEST_PROTOCOL)

    # if we know for sure the sim has been done, we can just load it. Since time-stepping only fills the Udata attribute
    # to the simulation object, "loading" a saved sim just means
    # 1) loading the pickle and
    # 2) adding the Udata attribute to our new sim
    def load(self):

        try:

            with open('joe_sim_archive/' + self.filename, 'rb') as inp:
                loaded_sim = pickle.load(inp)
                self.Udata = loaded_sim.Udata

        # due to old filenaming conventions where sometimes what is now the boolean "sponge_layer" was called
        # "absorbing_layer" or "abslayer", we want to allow the possibility of loading old files.
        # TODO: This is only for Adam George Morgan and should be deprecated in future releases/ once the relevant
        #  papers are published

        except:

            old_string = ('_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (self.length, self.T, self.N,
                                                                self.dt) + '_modelkw=' + self.model_kw + '_ICkw=' +
                            self.initial_state_kw + '_nonlinear=' + str(self.nonlinear) + '_abslayer='
                            + str(self.sponge_layer))

            old_filename = 'simdata' + old_string + '.pkl'

            with open('joe_sim_archive/' + old_filename, 'rb') as inp:
                loaded_sim = pickle.load(inp)
                self.Udata = loaded_sim.Udata


    # now put everything together: load if possible, but run if you gotta
    def load_or_run(self, method_kw='etdrk4', print_runtime=True, save=True, verbose=True):

        try:
            self.load()
            if verbose:
                print('Saved simulation found, loading saved data.')
            else:
                pass

        except:
            if verbose:
                print('No saved simulation found, running simulation.')
            else:
                pass
            self.run_sim(method_kw=method_kw, print_runtime=print_runtime)

            if save:
                self.save()
            else:
                pass

    # plot the initial state
    def plot_initial_condition(self, custom_ylim = None, dpi=100, color='xkcd:cerulean', fieldname='u', usetex=True,
                           show_figure=True, save_figure=False):

        x = clip_spongeless(self.x, self.sfrac)

        if self.t_ord == 1:
            u = self.initial_state

        elif self.t_ord == 2:
            u = self.initial_state[0,:]

        else:
            raise ValueError('t_ord must be 1 or 2.')

        u = clip_spongeless(u, self.sfrac)

        if usetex:

            try:
                ylabel = r'$u(x,t=0)$'.replace('u', str(fieldname))

                nice_plot(x, u, r'$x$', ylabel, custom_ylim=custom_ylim,
                          show_figure=show_figure, save_figure=save_figure, picname=self.ic_picname,
                          linestyle='solid',
                          color=color, usetex=True, dpi=dpi)

            except RuntimeError:  # catch a user error thinking they have tex when they don't
                usetex = False

        else:

            ylabel = r'u(x,t=0)'.replace('u', str(fieldname))

            nice_plot(x, u, r'x', ylabel, custom_ylim=custom_ylim, show_figure=show_figure,
                      save_figure=save_figure, picname=self.ic_picname,
                      linestyle='solid',
                      color=color, usetex=False, dpi=dpi)


    # create a Hovmoeller plot (filled contour plot in space-time) of the simulation.
    def hov_plot(self, umin=None, umax=None, dpi=100, cmap='cmo.haline', fieldname='u', usetex=True, show_figure=True, save_figure=False):
        nsteps = int(self.T / self.dt)
        times = np.linspace(0., self.T, num=1 + int(nsteps / self.ndump), endpoint=True)

        if self.t_ord == 1:
            u = self.Udata

        elif self.t_ord == 2:
            u = self.Udata[0, :, :]

        else:
            raise ValueError('t_ord can only be 1 or 2.')

        # add right endpoint to prevent a stripe from appearing in the pics
        x_end = np.append(self.x, 0.5 * self.length)
        x_end = clip_spongeless(x_end, self.sfrac)

        if self.complex:
            u_end = 1j*np.zeros((1 + int(self.T / (self.ndump * self.dt)), self.N + 1), dtype=float)

        else:

            u_end = np.zeros((1 + int(self.T / (self.ndump * self.dt)), self.N + 1), dtype=float)

        u_end[:, 0:self.N] = np.copy(u)

        u_end[:, -1] = np.copy(u[:, 0])

        u_end = clip_spongeless(u_end, self.sfrac)

        with spinner('Rendering Hovmoeller plot...'):

            # if the field is real, nothing to do
            if not self.complex:

                hov_plot(x_end, times, u_end, fieldname=fieldname, umin=umin, umax=umax, dpi=dpi, show_figure=show_figure, save_figure=save_figure,
                         picname=self.picname, cmap=cmap, usetex=usetex)

            # if the field is complex, we want to split it into its real and imaginary parts and plot these
            else:

                if usetex:

                    real_fieldname = r'\mathrm{Re}\left(u\right)'.replace('u', str(fieldname))
                    imag_fieldname = r'\mathrm{Im}\left(u\right)'.replace('u', str(fieldname))

                else:

                    real_fieldname = r'Re(u)'.replace('u', str(fieldname))
                    imag_fieldname = r'Im(u)'.replace('u', str(fieldname))

                hov_plot(x_end, times, np.real(u_end), fieldname=real_fieldname, umin=umin, umax=umax, dpi=dpi,
                         show_figure=show_figure, save_figure=save_figure,
                         picname=self.realpicname, cmap=cmap, usetex=usetex)

                hov_plot(x_end, times, np.imag(u_end), fieldname=imag_fieldname, umin=umin, umax=umax, dpi=dpi,
                         show_figure=show_figure, save_figure=save_figure,
                         picname= self.imagpicname, cmap=cmap, usetex=usetex)


    # as above, but just with the modulus of the soln
    def hov_plot_modulus(self, umin=None, umax=None, dpi=100, cmap='cmo.haline', fieldname='u', usetex=True, show_figure=True, save_figure=False):
        nsteps = int(self.T / self.dt)
        times = np.linspace(0., self.T, num=1 + int(nsteps / self.ndump), endpoint=True)

        if self.t_ord == 1:
            u = self.Udata

        elif self.t_ord == 2:
            u = self.Udata[0, :, :]

        else:
            raise ValueError('t_ord can only be 1 or 2.')

        # add right endpoint to prevent a stripe from appearing in the pics
        x_end = np.append(self.x, 0.5 * self.length)
        x_end = clip_spongeless(x_end, self.sfrac)

        if self.complex:

            u_end = 1j*np.zeros((1 + int(self.T / (self.ndump * self.dt)), self.N + 1), dtype=float)

        else:

            u_end = np.zeros((1 + int(self.T / (self.ndump * self.dt)), self.N + 1), dtype=float)

        u_end[:, 0:self.N] = np.copy(u)

        u_end[:, -1] = np.copy(u[:, 0])

        u_end = clip_spongeless(u_end, self.sfrac)

        u_end = np.absolute(u_end)

        with spinner('Rendering Hovmoeller plot of modulus...'):

            if usetex:

                mod_fieldname = r'\left|u\right|'.replace('u', str(fieldname))

            else:

                mod_fieldname = r'|u|'.replace('u', str(fieldname))

            hov_plot(x_end, times, u_end, fieldname=mod_fieldname, umin=umin, umax=umax, dpi=dpi, show_figure=show_figure, save_figure=save_figure,
                         picname=self.modpicname, cmap=cmap, usetex=usetex)


    # save a movie of the evolution of our solution.
    def save_movie(self, fps=200, fieldname='u', usetex=True, fieldcolor='xkcd:ocean green', dpi=100):

        if self.t_ord == 1:
            u = clip_spongeless(self.Udata, self.sfrac)

        elif self.t_ord == 2:
            u = clip_spongeless(self.Udata[0, :, :], self.sfrac)

        else:
            pass

        with spinner('Rendering movie...'):

            if not self.complex:
                save_movie(u, x=clip_spongeless(self.x, self.sfrac), length=self.length, dt=self.dt,
                           fieldname=fieldname, fps=fps, ndump=self.ndump, filename=self.moviename,
                           periodic=not self.sponge_layer, usetex=usetex, fieldcolor=fieldcolor, dpi=dpi)

            else:

                if usetex:

                    real_fieldname = r'\mathrm{Re}\left(u\right)'.replace('u', str(fieldname))
                    imag_fieldname = r'\mathrm{Im}\left(u\right)'.replace('u', str(fieldname))

                else:

                    real_fieldname = r'Re(u)'.replace('u', str(fieldname))
                    imag_fieldname = r'Im(u)'.replace('u', str(fieldname))

                save_movie(np.real(u), x=clip_spongeless(self.x, self.sfrac), length=self.length, dt=self.dt,
                           fieldname=real_fieldname, fps=fps, ndump=self.ndump, filename=self.realmoviename,
                           periodic=not self.sponge_layer, usetex=usetex, fieldcolor=fieldcolor, dpi=dpi)

                save_movie(np.imag(u), x=clip_spongeless(self.x, self.sfrac), length=self.length, dt=self.dt,
                           fieldname=imag_fieldname, fps=fps, ndump=self.ndump, filename=self.imagmoviename,
                           periodic=not self.sponge_layer, usetex=usetex, fieldcolor=fieldcolor, dpi=dpi)


    # save a movie the modulus
    def save_movie_modulus(self, fps=200, fieldname='u', usetex=True, fieldcolor='xkcd:ocean green', dpi=100):

        if self.t_ord == 1:
            u = clip_spongeless(self.Udata, self.sfrac)

        elif self.t_ord == 2:
            u = clip_spongeless(self.Udata[0, :, :], self.sfrac)

        else:
            pass

        u = np.absolute(u)

        with spinner('Rendering movie of modulus...'):

            if usetex:

                mod_fieldname = r'\left|u\right|'.replace('u', str(fieldname))

            else:

                mod_fieldname = r'|u|'.replace('u', str(fieldname))

            save_movie(u, x=clip_spongeless(self.x, self.sfrac), length=self.length, dt=self.dt,
                            fieldname=mod_fieldname, fps=fps, ndump=self.ndump, filename=self.modmoviename,
                            periodic=not self.sponge_layer, usetex=usetex, fieldcolor=fieldcolor, dpi=dpi)

    # save a movie of the evolution of our perturbation AND a nested movie of its power spectrum
    def save_combomovie(self, fps=200, fieldname='u', fieldcolor='xkcd:ocean green', speccolor='xkcd:dark orange', usetex=True, dpi=100):
        if self.t_ord == 1:
            u = clip_spongeless(self.Udata, self.sfrac)

        elif self.t_ord == 2:
            u = clip_spongeless(self.Udata[0, :, :], self.sfrac)

        with spinner('Rendering combo movie...'):

            # for a real field there's nothing to do....
            if not self.complex:

                save_combomovie(u,  x=clip_spongeless(self.x, self.sfrac), length=self.length, dt=self.dt,
                                fieldname=fieldname, fps=fps, fieldcolor=fieldcolor,
                                speccolor=speccolor, ndump=self.ndump, filename=self.combomoviename,
                                periodic=not self.sponge_layer, usetex=usetex, dpi=dpi)

            # for a complex field just plot the modulus and the power spectrum
            else:

                u = np.absolute(u)

                if usetex:

                    mod_fieldname = r'\left|u\right|'.replace('u', str(fieldname))

                else:

                    mod_fieldname = r'|u|'.replace('u', str(fieldname))

                save_combomovie(u, x=clip_spongeless(self.x, self.sfrac), length=self.length, dt=self.dt,
                                fieldname=mod_fieldname, fps=fps, fieldcolor=fieldcolor,
                                speccolor=speccolor, ndump=self.ndump, filename=self.combomoviename,
                                periodic=not self.sponge_layer, usetex=usetex, dpi=dpi)

    # obtain first moment of the system
    def get_fm(self):

        length = self.length
        N = self.N
        u = self.Udata

        fm = (1./length) * integrate(u, length, N) # take mean

        fm_error = np.abs(fm[1:]-fm[0])

        self.fm = fm
        self.fm_error = fm_error

    # obtain second moment
    def get_sm(self):

        length = self.length
        N = self.N
        u = self.Udata

        if self.fm.any() is None:

            self.get_fm()

        fm = self.fm

        sm = (1./length) * integrate(u**2, length, N)
        sm -= fm**2

        sm_error = np.abs(sm[1:]-sm[0])

        self.sm = sm
        self.sm_error = sm_error

# a function that performs a refinement study based on Richardson extrapolation for error estimation. Very
# useful for quickly checking accuracy. It's here because it doesn't have a better home right now, and it
# *almost* takes simulation objects in as input.
def do_refinement_study(model, initial_state, length, T, Ns, dts, method_kw='etdrk4', bc='periodic', sponge_params=None,
                        show_figure=True, save_figure=False, usetex=True, dpi=400, fit_min=3, fit_max=7):

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False

    Ns = Ns.astype(int)
    num_Ns = np.size(Ns)
    num_dts = np.size(dts)

    # initialize outputs

    errors = np.zeros([num_Ns, num_dts], dtype=float)

    cnt = 0

    #start = time.time()
    with spinner('Performing refinement study...'):
        for k in np.arange(0, num_Ns):

            N = Ns[k]

            # do simulation at the worst order (largest time step) first
            rough_st_grid = {'length':length, 'T':T, 'N':N, 'dt':dts[0]}
            rough_sim = simulation(rough_st_grid, model, initial_state, bc=bc, sponge_params=sponge_params)

            rough_sim.load_or_run(method_kw=method_kw, save=True, print_runtime=False, verbose=False)

            for dt in dts:

                fine_st_grid = {'length': length, 'T': T, 'N': N, 'dt': 0.5*dt}
                fine_sim = simulation(fine_st_grid, model, initial_state, bc=bc, sponge_params=sponge_params)

                fine_sim.load_or_run(method_kw=method_kw, save=True, print_runtime=False, verbose=False)

                rough_Udata = rough_sim.Udata

                fine_Udata = fine_sim.Udata

                # use fine sim and rough sim at last time step to get Richardson error estimate

                if fine_sim.t_ord == 1:

                    diff = clip_spongeless(rough_Udata[-1, :] - fine_Udata[-1, :], fine_sim.sfrac)

                elif fine_sim.t_ord == 2:

                    diff = clip_spongeless(rough_Udata[0, -1, :] - fine_Udata[0, -1, :], fine_sim.sfrac)

                else:

                    raise ValueError('Order of temporal derivatives must be 1 or 2')

                errors[k, cnt] = (1. / ((2 ** 4) - 1)) * np.amax(np.absolute(diff))

                rough_sim = fine_sim  # redefine for efficiency... only works bcz we refine dt in powers of 1/2

                cnt += 1

            cnt = 0  # reinit the counter

    #end = time.time()
    #runtime = end - start
    #print('Runtime for accuracy tests = %.4f' % runtime + ' s')

    # now we produce a plot of the errors using an awkward but functioning purpose-built plotting fnc
    plot_refinement_study(model, initial_state, length, T, Ns, dts, errors, method_kw=method_kw, bc=bc,
                        show_figure=show_figure, save_figure=save_figure, usetex=usetex, dpi=dpi)

    # estimate the slope of particular error curves if you want. Needs a bit of by-hand tweaking (controlled by the
    # inputs fit_min, fit_max) bcz for small enough dt we can get level-off or rounding error domination in the error
    # curve, destroying the linear trend after a certain threshold

    params = np.polyfit(np.log10(dts[fit_min:fit_max+1]), np.log10(errors[-1, fit_min:fit_max+1]), 1)
    slope = params[0]
    print('Estimated Slope of Error Line at N = %i' % Ns[-1] + ' is slope = %.3f' % slope)

    return None

def do_refinement_study_alt(model, initial_state, length, T, Ns, dts, benchmark_sim, method_kw='etdrk4', bc='periodic', sponge_params=None, show_figure=True, save_figure=False, usetex=True,
                        fit_min=3, fit_max=7):

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False

    Ns = Ns.astype(int)
    num_Ns = np.size(Ns)
    num_dts = np.size(dts)

    # initialize outputs

    errors = np.zeros([num_Ns, num_dts], dtype=float)

    cnt = 0

    benchmark_sim.load_or_run(method_kw=method_kw, save=True, print_runtime=False, verbose=False)

    #start = time.time()
    with spinner('Performing refinement study...'):
        for k in np.arange(0, num_Ns):

            N = Ns[k]

            for dt in dts:

                stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
                rough_sim = simulation(stgrid, model, initial_state, bc=bc, sponge_params=sponge_params)

                rough_sim.load_or_run(method_kw=method_kw, save=True, print_runtime=False, verbose=False)

                rough_Udata = rough_sim.Udata

                benchmark_Udata = benchmark_sim.Udata

                # use fine sim and rough sim at last time step to get Richardson error estimate

                diff = clip_spongeless(rough_Udata[-1, :]-benchmark_Udata[-1, :], benchmark_sim.sfrac)

                errors[k, cnt] = np.amax(np.abs(diff))

                cnt += 1

            cnt = 0  # reinit the counter

    #end = time.time()
    #runtime = end - start
    #print('Runtime for accuracy tests = %.4f' % runtime + ' s')

    # now we produce a plot of the errors
    fig, ax = plt.subplots()

    dts = 0.5 * dts

    # define the cycler
    my_cycler = (
                plt.cycler(color=['xkcd:slate', 'xkcd:raspberry', 'xkcd:goldenrod', 'xkcd:deep green'])
                + plt.cycler(lw=[3.5, 3, 2.5, 2])
                + plt.cycler(linestyle=['dotted', 'dashed', 'solid', 'dashdot'])
                + plt.cycler(marker=['v', '*', 'o', 'P'])
                + plt.cycler(markersize=[8, 12, 8, 8])
    )

    ax.set_prop_cycle(my_cycler)

    for m in range(0, num_Ns):
        if usetex:
            plt.loglog(dts, errors[m, :], label=r'$N = z$'.replace('z', str(Ns[m])))
        # ^ an awesome trick from
        # https://stackoverflow.com/questions/33786332/matplotlib-using-variables-in-latex-expressions
        # was used to get the labels working as above
        else:
            plt.loglog(dts, errors[m, :], label='N = z'.replace('z', str(Ns[m])))

    ax.legend(fontsize=16)

    if usetex:
        plt.xlabel(r"$\Delta t$", fontsize=26, color='k')
        plt.ylabel(r"Absolute Error", fontsize=26, color='k')
    else:
        plt.xlabel("Δt", fontsize=26, color='k')
        plt.ylabel("Absolute Error", fontsize=26, color='k')

    plt.tick_params(axis='x', which='both', top='off', color='k')
    plt.xticks(fontsize=16, rotation=0, color='k')
    plt.tick_params(axis='y', which='both', right='off', color='k')
    plt.yticks(fontsize=16, rotation=0, color='k')

    plt.tight_layout()

    if save_figure is True:

        # add the folder "joe_visuals" to our path
        my_path = os.path.join("joe_visuals")

        # first, if the folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        # and now we can save the fig
        if bc == 'sponge_layer':
            sponge_layer = True
        elif bc == 'periodic':
            sponge_layer = False

        my_string = ('_length=%.1f_T=%.1f' % (
        length, T) + '_modelkw=' + model.model_kw + '_ICkw=' + initial_state.initial_state_kw + '_method_kw='
                     + method_kw + '_nonlinear=' + str(model.nonlinear) + '_abslayer=' + str(sponge_layer))

        picname = 'refinement_study' + my_string + '.png'
        plt.savefig('joe_visuals/' + picname, bbox_inches='tight', dpi=400)

    else:

        pass

    if show_figure is True:

        plt.show()

    else:

        pass

    plt.close()

    # estimate the slope of particular error curves if you want. Needs a bit of by-hand tweaking (controlled by the
    # inputs fit_min, fit_max) bcz for small enough dt we can get level-off or rounding error domination in the error
    # curve, destroying the linear trend after a certain threshold

    params = np.polyfit(np.log10(dts[fit_min:fit_max+1]), np.log10(errors[-1, fit_min:fit_max+1]), 1)
    slope = params[0]
    print('Estimated Slope of Error Line at N = %i' % Ns[-1] + ' is slope = %.3f' % slope)

    return None



import numpy as np
np.random.seed(32)

from scipy.fft import rfft, irfft, rfftfreq

import joe_lab.joe as joe
from joe_lab.visualization import nice_plot
from joe_lab.utils import integrate

def sample_one_phase(length):
    return np.random.uniform(low=-0.5 * length, high=0.5 * length)
def get_sample_phases(m, length, min_dist=20.):
    x = np.zeros(m)

    x[0] = sample_one_phase(length)

    for k in range(1, m):

        draw_new_sample = True

        while draw_new_sample:

            y = sample_one_phase(length)

            ds = np.array([np.abs(y - xx) >= min_dist for xx in x[0:k]])

            bdry_ds = np.array([np.abs(y - xx) >= 0.5*min_dist for xx in np.array([-0.5*length, 0.5*length])])

            if ds.all() and bdry_ds.all():

                draw_new_sample = False

        x[k] = y

    return x

S = 1.
delta = 2.

a_crit = -3./(delta*S) # amplitude below which we see negative solitary waves

def bbm_solitary_wave(x,c=1.):
    out = np.zeros_like(x, dtype=float)
    xmax = 180
    out[abs(x) > xmax] = 0.
    xx = x[abs(x) <= xmax]
    out[abs(x) <= xmax] = (3. / S) * c * (np.cosh(0.5 * np.sqrt(c / (1. + delta * c)) * (xx)) ** -2)
    return out

def my_symbol(k):
    return 1j * (k ** 3) / (1. + delta*k ** 2)

def my_fourier_forcing(V,k,x,nonlinear=True):
    p = 1.

    out = -S * float(nonlinear) * (1. / (p + 1.)) * (1j * k / (1. + delta*k ** 2) )*  rfft(irfft(V) ** (p + 1))

    return out

def soliton_gas_ic(x,m,length):

    out = 0.

    #phases = get_sample_phases(m, length, min_dist=0.033*length)
    phases = np.linspace(-0.5 * length + 30, 0.5 * length - 30, num=m, endpoint=True)

    amps = np.random.normal(loc=1., scale=np.sqrt(0.2), size=m) #np.random.uniform(low=.5, high=1.4, size=m)

    # mm = int(0.5*m)

    # amps_plus = np.random.uniform(low=.5, high=1.4, size=mm)

    # amps_minus = np.random.uniform(low=a_crit-0.5, high=a_crit-0.3, size=mm)

    # amps = np.concatenate((amps_plus, amps_minus))

    cs = (S/3.)*amps

    np.random.shuffle(cs)

    for k in range(0,m):

        out += bbm_solitary_wave(x-phases[k], c=cs[k])

    return out

# params are set to be at 1/4 of what DP2014 had in terms of spatial grid, so we have 1/4 of the solitons
# as well
length, T, N, dt = 2.*1395., 2., 2**13, 4e-3
m = 50 # number of solitons in the gas

ndump = int(1./dt)

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
my_model = joe.model('bbm_dp2014', 1, my_symbol, my_fourier_forcing, nonlinear=True)
my_initial_state = joe.initial_state('soliton_gas_gaussian', lambda x: soliton_gas_ic(x, m, length))

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=ndump)

#my_sim.plot_initial_condition(show_figure=True, save_figure=False)

my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=False)
my_sim.hov_plot(cmap = 'cmo.haline', usetex=True, save_figure=False, show_figure=True)
#my_sim.save_movie(dpi=200, fps=200, usetex=False, fieldcolor='xkcd:cerulean', fieldname='u')


my_sim.get_fm()
fm_error = np.amax(my_sim.fm_error)
print(fm_error)


# compute the BBM energy (Sobolev H^1 norm)
def energy(u):
    # get wavenumbers for the grid of S^1 with N samples
    k = 2. * np.pi * N * rfftfreq(N) / length

    spring = irfft(1j * k * rfft(u)) ** 2

    out = integrate(u**2 + delta*spring, length, N)

    return out

# get the energies associated to each time
times = np.linspace(0., T, num=1 + int(T / (dt*my_sim.ndump)), endpoint=True)
Udata = my_sim.Udata
E = energy(Udata)
E_error = E-E[0]
E_error_rel = E_error/E[0]
print('Max absolute error in energy = ', np.amax(np.abs(E_error)))
print('Max relative error in energy =', np.amax(np.abs(E_error_rel)))

# draw the figure
dpi = 400
picname = 'bbm_dp2014_st_gaussian_energy_test_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '_ICkw=' + my_sim.initial_state_kw  + '.png'
nice_plot(times, E_error, r'$t$', r'Error in Energy', dpi=dpi, show_figure=False,
          save_figure=True, picname=picname, linestyle='solid', color='xkcd:blueberry', usetex=True)

picname = 'bbm_dp2014_st_gaussian_energy_test_rel_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '_ICkw=' + my_sim.initial_state_kw  + '.png'
nice_plot(times, E_error, r'$t$', r'Relative Error in Energy', dpi=dpi, show_figure=False,
          save_figure=True, picname=picname, linestyle='solid', color='xkcd:blueberry', usetex=True)



import numpy as np

from scipy.fft import fft, ifft

import joe_lab.joe as joe

g = 1.

def my_symbol(k):
    return -1j*k**2

def my_fourier_forcing(V,k,x,nonlinear=True):

    u = ifft(V)

    out =  fft(float(nonlinear)*1j*g*u*np.absolute(u)**2)

    return out

model_kw = 'focusing_nls'

# if the soln is complex, this must be told to the program explicitly
my_model = joe.model(model_kw, 1, my_symbol, my_fourier_forcing, nonlinear=True, complex=True)

def plane_wave(x):
    return np.ones_like(x) + 1j*np.zeros_like(x)

def nls_soliton(x,a=1., c=1.):

    out = 1j*np.zeros_like(x)

    xmax = 180
    out[abs(x) > xmax] = 0.
    out[abs(x) <= xmax] = np.sqrt(2.*a)*np.exp(1j*0.5*c*x[abs(x) <= xmax])/np.cosh(np.sqrt(a)*x[abs(x) <= xmax])

    return out

my_initial_state = joe.initial_state('nls_soliton', nls_soliton)
#initial_state('plane_wave', plane_wave) #

length, T, N, dt = 400., 400., 2**10, 1e-2
stgrid = {'length':length, 'T':T, 'N':N, 'dt':dt}

l_endpt = -length * 0.5
r_endpt = l_endpt + 4e-4 * length
width = (2 ** -6) * length / 100.
sponge_params = {'l_endpt': l_endpt, 'r_endpt': r_endpt,
                 'width': width, 'expdamp_freq': 2,
                 'damping_amplitude': 30.,
                 'splitting_method_kw': 'naive',
                 'spongeless_frac': .5}  # this is the fraction of the middle of the spatial domain to keep in the plots

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='sponge_layer', sponge_params=sponge_params, ndump=20)
#my_sim = simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=20)

my_sim.load_or_run(method_kw='etdrk4', save=False)

#my_sim.hov_plot(show_figure=False, save_figure=True, usetex=False, cmap='plasma')
my_sim.hov_plot_modulus(show_figure=True, save_figure=True, usetex=False, cmap='RdPu')
#my_sim.save_movie(fps=100, usetex=False, fieldcolor='xkcd:heliotrope')
#my_sim.save_movie_modulus(fps=100, usetex=False, fieldcolor='xkcd:barney purple')
#my_sim.save_combomovie(fps=100, usetex=False, fieldcolor='xkcd:barney purple')

"""
nmin, nmax = 2, 12
Ns = np.array([2**8, 2**9, 2**10])
dts = np.flip(np.logspace(-nmax, -nmin, num=nmax-nmin+1, base=2.))

joe.do_refinement_study(my_model, my_initial_state, length, T, Ns, dts, bc='periodic', method_kw='etdrk4',
                    show_figure=True, save_figure=True, usetex=True, fit_min=3, fit_max=9)
"""



import numpy as np
from scipy.fft import rfft, irfft

import joe_lab.joe as joe
import joe_lab.initial_states as joe_initial_states

length, T, N, dt = 30., 30., 2**9, 1e-2

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}

model_kw = 'kn_null_wave'

def symbol(k):
    return -k**2

def fourier_forcing(V, k, x, nonlinear=True):
    if int(V.size - 2) == x.size:

        pass

    else:

        raise TypeError("The array V must be 2+(size of the array x).")

    N = int(V.size - 2)

    V = np.reshape(V, (N + 2,))

    NN = int(0.5 * N) + 1

    u = irfft(V[0:NN])  # only ifft first N entries of V because of storage conventions

    ux = irfft(1j * k * V[0:NN])

    v = irfft(V[NN:N+2])

    spatial_forcing = float(nonlinear) * (ux ** 2 - v ** 2)

    out = 1j * np.zeros(N + 2, dtype=float)
    out[NN:] = rfft(spatial_forcing)

    return out

my_model = joe.model(model_kw, 2, symbol, fourier_forcing, nonlinear=True)

my_initial_state = joe_initial_states.builtin_initial_state('kdv_soliton')

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=20)

my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=True)

# produce plots and movies
my_sim.hov_plot(cmap='cmo.matter', fieldname='u', show_figure=True, save_figure=True, usetex=True)
#my_sim.save_movie(dpi=200, fps=45, usetex=False, fieldcolor='xkcd:deep magenta', fieldname='u')



# joe code for reproducing the experiments and (most) statistical results from a paper of Didenkulova
# https://www.sciencedirect.com/science/article/abs/pii/S0167278918305657

import time

import numpy as np
from scipy.signal import argrelmin, argrelmax # for postprocessing: amplitude hist and cdf
from scipy.stats import ecdf # again for postprocessing

import joe_lab.joe as joe
import joe_lab.models as joe_models
from joe_lab.utils import integrate
from joe_lab.visualization import spinner, nice_plot, nice_multiplot, nice_hist

from joblib import Parallel, delayed

##############################################################################################
########################## STAGE 1: SIMULATION OF THE ENSEMBLE ###############################
##############################################################################################

# fix basic params
num_samples = 50
num_waves = 30

ndump = 500

# get stgrid
length, T, N, dt = 600., 200., 2 ** 13, 2e-5
stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}

# get model
my_model = joe_models.builtin_model('gardner', nonlinear=True)

# get initial state
# note that, for large domains, cosh(x near endpts) may encounter overflow, so it's good to just manually set
# the tails of the wave to zero to avoid these concerns (see also the great discussion here
# https://stackoverflow.com/questions/31889801/overflow-in-numpy-cosh-function
# which has inspired my approach below)
def gardner_soliton(x, c=1., p=1.):
    out = np.zeros_like(x, dtype=float)
    xmax = 180
    out[abs(x) > xmax] = 0.
    out[abs(x) <= xmax] = c / (-1. + p * np.sqrt(1. + c) * np.cosh(np.sqrt(c) * x[abs(x) <= xmax]))
    return out

def soliton_gas_ic(x, m):
    out = 0.

    phases = np.linspace(-0.5 * length + 10, 0.5 * length - 10, num=m, endpoint=True)

    mm = int(0.5 * m)

    amps_plus = np.random.uniform(low=2.3, high=3., size=mm)

    cs_plus = (amps_plus - 1.) ** 2 - 1.  # using the Gardner NL dispersion relation

    amps_minus = -1. * np.random.uniform(low=.1, high=3., size=mm)

    cs_minus = (np.abs(amps_minus) + 1.) ** 2 - 1.  # using the Gardner NL dispersion relation

    cs = np.concatenate((cs_plus, cs_minus))

    ps = np.ones(m, dtype=float)

    ps[mm:] *= -1

    z = np.zeros((m, 2), dtype=float)

    z[:, 0] = cs

    z[:, 1] = ps

    np.random.shuffle(z)

    for k in range(0, m):
        out += gardner_soliton(x - phases[k], c=z[k, 0], p=z[k, 1])

    return out

# define a function that takes in a sample number and does a sim
def sample_st(sample):
    ic_string = 'st_sample_' + str(sample)

    my_initial_state = joe.initial_state(ic_string, lambda x: soliton_gas_ic(x, num_waves))

    my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=ndump)
    my_sim.load_or_run(print_runtime=False, verbose=False, save=True)

    my_sim.get_fm()
    my_sim.get_sm()
    fm_error = np.amax(my_sim.fm_error)
    sm_error = np.amax(my_sim.sm_error)
    return np.array([fm_error, sm_error])

start = time.time()

with spinner('Simulating Gardner soliton turbulence...'):
    errors = Parallel(n_jobs=2)(delayed(sample_st)(sample) for sample in range(1, num_samples+1))
    # depending on your machine, n_jobs or the number of batches you split your ensemble into may
    # need to be altered!
    errors = np.array(errors)
    fm_errors = errors[:, 0]
    sm_errors = errors[:, 1]

end = time.time()
runtime = end - start
print('Runtime for Gardner soliton turbulence simulation = %.4f' % runtime + ' s')

print('Maximum error in first moment over ensemble = ', np.amax(fm_errors))
print('Maximum error in second moment over ensemble = ', np.amax(sm_errors))

import sys
sys.exit()

##############################################################################################
########################## STAGE 2: PREP FOR POSTPROCESSING (HELPER FNCS ETC.) ###############
##############################################################################################

num_timesteps = 1+int(T/(dt*ndump))
tt = ndump*dt*np.arange(0, num_timesteps)

# code for computing higher moments given field and its first two moments
def get_higher_moments(u,fm,sm):

    v = (u-fm)**3

    skew = (1./length)*integrate(v, length, N)
    skew /= sm**1.5

    v = (u-fm)**4

    kurt = (1./length)*integrate(v, length, N)
    kurt /= sm**2

    return skew, kurt

# code for computing CDF of unif distribution on [a,b]
def unif_cdf(x, a=-1, b=1.):
    out = np.zeros_like(x, dtype=float)
    out[x>a] = (x[x>a] - a)/(b-a)
    out[x>=b] = 1.
    return out

##############################################################################################
########################## STAGE 3: POSTPROCESSING ###########################################
##############################################################################################

# init storage
skew_store = np.zeros((num_samples, num_timesteps), dtype=float)
kurt_store = np.zeros((num_samples, num_timesteps), dtype=float)
max_storage = np.zeros((num_samples, num_timesteps), dtype=float)
min_storage = np.zeros((num_samples, num_timesteps), dtype=float)
pos_amp = []
neg_amp = []

for sample in range(1, num_samples+1):

    ic_string = 'st_sample_' + str(sample)
    my_initial_state = initial_state(ic_string, lambda x: soliton_gas_ic(x, num_waves))
    my_sim = simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=ndump)
    my_sim.load()

    """
    # if desired, save an example initial state plot and Hovmoeller plot for one sample
    if sample == 1:
        my_sim.plot_initial_condition(color='xkcd:cerulean', usetex=True, show_figure=False, save_figure=True)

        my_sim.hov_plot(umin=-3., umax=3., dpi=600, usetex=True, save_figure=True, show_figure=False, cmap='cmo.thermal')
        # after a lot of experimenting I really think the thermal colormap is the right way to go
        # for Gardner, where the really thin antisolitons need to stand out as strongly as possible
    """

    u = my_sim.Udata

    # moments
    my_sim.get_fm()
    fm = my_sim.fm[0] #TODO: better to put in "actual" or imperfect first/second moments
                      # into higher moment computations? Honestly the precision of the moments is so good it barely matters
    my_sim.get_sm()
    sm = my_sim.sm[0]

    skew, kurt = get_higher_moments(my_sim.Udata, fm, sm)
    skew_store[sample-1, :] = skew
    kurt_store[sample-1, :] = kurt

    # +/- amplitude asymmetry plot
    umax = np.amax(u, axis=1)
    umin = np.amin(u, axis=1)

    max_storage[sample-1] = umax
    min_storage[sample-1] = umin

    # amplitude histograms and CDF

    # need to set a "mixing time" after which we assume the "gas" is well-mixed and we're OK to start doing some data
    # analysis. Honestly 0.25*T could even be a bit large!
    mixing_time = 0.25 * T

    mixing_index = int(mixing_time / (ndump * dt))

    # only count stuff after mixing time
    u = u[mixing_index:, :]

    delta_pos = 2e0  # threshold: ignore + waves below this amplitude
    delta_min = -1e-1  # same but for - waves

    pos_part = u[u >= delta_pos]
    max = argrelmax(pos_part, axis=0)
    pos_amp.extend(pos_part[max])

    neg_part = u[u <= delta_min]
    min = argrelmin(neg_part, axis=0)
    neg_amp.extend(neg_part[min])

##############################################################################################
########################## STAGE 4: PRODUCE POSTPROCESSING PLOTS #############################
##############################################################################################

dpi = 600

# Plot skew and kurtosis averaged over the ensemble
picname = 'gardner_st_skew_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'
nice_plot(tt, np.mean(skew_store, axis=0), r"$t$", r"Skewness", custom_ylim=None,
          dpi=dpi,show_figure=False, save_figure=True, picname=picname, linestyle='solid', color='xkcd:pumpkin',
          usetex=True)

picname = 'gardner_st_kurt_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'
nice_plot(tt, np.mean(kurt_store, axis=0), r"$t$", r"Kurtosis", custom_ylim=None, dpi=dpi,
          show_figure=False, save_figure=True, picname=picname, linestyle='solid',
          color='xkcd:pinky purple', usetex=True)

# Plot +/- amplitudes (max/min over ensemble resp) to illustrate asymmetry between polarities
picname = 'gardner_st_Amps_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'
nice_multiplot([tt,tt], [np.max(max_storage, axis=0), np.min(min_storage, axis=0)],
               r"$t$", r"Min/Max Amplitudes",
               curvelabels = ['Max', 'Min'],
               linestyles = ['dotted', 'dashed'], colors = ['xkcd:deep pink', 'xkcd:teal green'], linewidths = [2, 2],
               custom_ylim=[-6,4], dpi = dpi, show_figure = False, save_figure = True, picname = picname, usetex=True)

# Plot histograms of amplitudes

# + amplitudes
picname = 'gardner_st_hist+_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'
nice_hist(pos_amp, r'$A_{+}$', dpi=dpi, show_figure=False, save_figure=True, picname=picname,
              color='xkcd:deep pink', usetex=True)

# - amplitudes
picname = 'gardner_st_hist-_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'
nice_hist(neg_amp, r'$A_{-}$', dpi=dpi, show_figure=False, save_figure=True, picname=picname,
              color='xkcd:teal green', usetex=True)

# Plot amplitude CDFs

# + amplitudes
pos_cdf = ecdf(pos_amp)
xx = np.linspace(2,4, num=600)

picname = 'gardner_st_cdf+_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'
nice_multiplot([xx,xx], [pos_cdf.cdf.evaluate(xx), unif_cdf(xx, a=2.3, b=3)],
               r"$A_{+}$", r"$F\left(A_{+}\right)$",
               curvelabels = ['Observed', 'Uniform'],
               linestyles = ['solid', 'dashed'], colors = ['xkcd:deep pink', 'xkcd:blueberry'],
               linewidths = [3, 2], custom_ylim=[-0.02,1.02],
               dpi = dpi, show_figure = False, save_figure = True, picname = picname, usetex=True)

# - amplitudes
neg_cdf = ecdf(neg_amp)
xx = np.linspace(-6,0, num=600)

picname = 'gardner_st_cdf-_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'
nice_multiplot([xx,xx], [neg_cdf.cdf.evaluate(xx), unif_cdf(xx, a=-3., b=-0.1)],
               r"$A_{-}$", r"$F\left(A_{-}\right)$",
               curvelabels = ['Observed', 'Uniform'],
               linestyles = ['solid', 'dashed'], colors = ['xkcd:teal green', 'xkcd:blueberry'],
               linewidths = [3, 2], custom_ylim=[-0.02,1.02],
               dpi = dpi, show_figure = False, save_figure = True, picname = picname, usetex=True)

"""
#####################################################################################
##### Old code for checking to see if argrelmax/argrelmin are working correctly #####
#####################################################################################

import matplotlib.pyplot as plt
uu = u[-1,:]
plt.plot(my_sim.x, uu)

    
pos_part = uu[u[-1,:]>delta] # threshold to avoid false local maxes/exclude very small waves
max = argrelmax(pos_part) # local maximizers (up to threshold)
amp = pos_part[max] # actual + wave amplitudes

# draw for validation
x_plt = []
for a in amp:
    x_plt.append(np.where(uu == a))
x_plt = -0.5*length+(length/N)*np.array(x_plt)

plt.scatter(x_plt, amp, marker='o', color='xkcd:deep pink')

neg_part = uu[u[-1, :] < -delta]
min = argrelmin(neg_part) # local minimizers (up to threshold)
amp = neg_part[min]  # actual - wave amplitudes

# draw for validation
x_plt = []
for a in amp:
    x_plt.append(np.where(uu == a))
x_plt = -0.5*length+(length/N)*np.array(x_plt)

plt.scatter(x_plt, amp, marker='v', color='xkcd:teal green')
    
plt.show()
"""



import numpy as np

np.random.seed(32)

import joe_lab.joe as joe
import joe_lab.models as joe_models


def kdv_soliton(x, c=1.):
    return 0.5 * c * (np.cosh(0.5 * np.sqrt(c) * (x)) ** -2)


length, T, N, dt = 600., 2., 2 ** 12, 1e-4
m = 30  # number of solitons in the gas

def soliton_gas_ic(x, m):
    out = 0.

    phases = np.linspace(-0.5*length+10, 0.5*length-10, num=m, endpoint=True)

    amps = np.random.uniform(low=1., high=2., size=m)

    cs = 2. * amps

    for k in range(0, m):
        out += kdv_soliton(x - phases[k], c=cs[k])

    return out

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
my_model = joe_models.builtin_model('kdv', nonlinear=True)
my_initial_state = joe.initial_state('soliton_gas', lambda x: soliton_gas_ic(x, m))

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=200)

my_sim.load_or_run(print_runtime=True, save=True)
my_sim.hov_plot(dpi=600, usetex=True, save_figure=True, show_figure=True)
#my_sim.save_movie(dpi=200, fps=200, usetex=False, fieldcolor='xkcd:cerulean', fieldname='u')

nsteps = int(T / (200 * dt))
t = np.linspace(0, T, num=nsteps + 1, endpoint=True)

my_sim.get_fm()
my_sim.get_sm()
print(np.amax(my_sim.fm_error))
print(np.amax(my_sim.sm_error))



import numpy as np

import joe_lab.joe as joe
import joe_lab.models as joe_models
import joe_lab.initial_states as joe_initial_states
from joe_lab.visualization import nice_multiplot
from joe_lab.utils import integrate


length, T, N, dt = 240.,2e4, 2**9, 1e-2

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
my_model = joe_models.builtin_model('phi4pert', nonlinear=True)
my_initial_state = joe_initial_states.builtin_initial_state('internal_mode')

# we need a sponge layer here, which requires parameter tuning to get right.

l_endpt = -0.5*length  + 0.5 * length * 0.05
r_endpt = l_endpt + 0.05 * length
width = (2 ** -4) * length / 100.
sponge_params = {'l_endpt': l_endpt, 'r_endpt': r_endpt,
                 'width': width, 'expdamp_freq': 1e3,
                 'damping_amplitude': 10.,
                 'spongeless_frac': 0.5}  # this is the fraction of the middle of the spatial domain to keep in the plots

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='sponge_layer', sponge_params=sponge_params, ndump=20)

my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=True)

# do data analysis on the amplitudes

# helper funcs first

from joe_lab.sponge_layer import clip_spongeless

def internal_mode(x):

    out = np.sinh(x / np.sqrt(2)) * (np.cosh(x / np.sqrt(2))) ** -2

    return out

# compute the L2 product of two real arrays via the FFT (sampled funcs)
def L2prod(u1, u2, length=length, N=N):

    M1 = u1.size

    M2 = u2.size

    if M1 == M2:

        pass

    else:

        raise TypeError("Arrays u1, u2 must have the same length")

    return integrate(u1*u2, length, N) 


# take in a simulation object (where the run has already been completed) and computes the amplitude of the internal mode
#
def amplitude(sim):

    sfrac = sim.sponge_params['spongeless_frac']

    slength, sN = sfrac*sim.length, sfrac*sim.N

    x = clip_spongeless(sim.x, sfrac)

    im = internal_mode(x)

    downstairs = L2prod(im, im, slength, sN)

    u = clip_spongeless(sim.Udata[0, :, :], sfrac)

    Nt, Nx = np.shape(u)

    amplitude = np.zeros(Nt, dtype=float)

    for n in np.arange(0, Nt):

        amplitude[n] = L2prod(u[n,:], im, slength, sN)/downstairs

    return amplitude

# now do the actual curve-fitting stuff

from scipy.optimize import curve_fit
from scipy.signal import argrelmax

nsteps = int(T / (my_sim.ndump*dt))
t = np.linspace(0, T, num=1 + nsteps, endpoint=True)

# get training data
A_simulated = amplitude(my_sim)
aa = -A_simulated # I empirically found the negative amplitudes are slightly higher, so I only use
# these for getting the training data for our fit

indices_train = argrelmax(aa)
A_train = aa[indices_train]

indices_train = np.array(indices_train)
t_train = dt*my_sim.ndump*indices_train.flatten()

# fit our own decay curve
def modelcurve(t, A, B, p):
    out = A * ((1. + B*t) ** p)
    return out

bounds = [np.array([0., 0., -1.]), np.array([1.,2.,0.1])]
params_opt, params_cov = curve_fit(modelcurve, t_train, A_train, p0=[0.1, 0.01, -0.5], maxfev=int(3e4), bounds=bounds)

print('Estimated L^2 error in fit = ', np.linalg.norm(A_train-modelcurve(t_train, *params_opt)))

A_my_model = modelcurve(t, *params_opt)

print('Optimal model parameters = ', params_opt)
print('Estimated error in parameters = ', np.sqrt(np.diag(params_cov)))

picname = 'phi4pert_amplitude_fit_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'

# and plot the results
nice_multiplot([t,t,t], [A_simulated, A_my_model, -A_my_model], r"$t$", r"$a(t)$",
               curvelabels = ['Computed', 'Empirical Fit', None],
               linestyles = ['solid', 'dashed', 'dashed'], colors = ['xkcd:teal', 'xkcd:pumpkin', 'xkcd:pumpkin'],
               linewidths = [0.8, 2.2, 2.2],
               dpi = 100, show_figure = True,
               save_figure = True, picname = picname, usetex = True)

# you may also wish to compare to the Delort-Masmoudi bound with the following code...

"""
# fit Delort-Masmoudi decay curve if desired
def DM_modelcurve(t, A):
    out = A * 0.1 / np.sqrt(1. + 0.01 * t)
    return out

DM_params_opt, DM_params_cov = curve_fit(DM_modelcurve, t_train, A_train, p0=[8.], maxfev=int(3e4))
# TODO: why do we need to take such a small tail to get DM picture looking good/persisting up to T=1e4?
A_DM = DM_modelcurve(t, *DM_params_opt)

print(DM_params_opt)
print(np.sqrt(np.diag(DM_params_cov)))  # estimated errors in parameter values
"""

# but generally I find the following empirical Delort-Masmoudi decay curve is much better asymptotically:
A_DM = params_opt[0]*(params_opt[1]**params_opt[2]) * 0.1 * ((1. + 0.01*t) ** -0.5)

picname = 'phi4pert_DM_comparison_length=%.1f_T=%.1f_N=%.1f_dt=%.6f' % (length, T, N, dt) + '.png'

nice_multiplot([t,t], [A_my_model, A_DM], r"$t$", r"$a(t)$",
               curvelabels = ['Empirical Fit', 'Empirical DM Bound'],
               linestyles = ['dashed', 'solid'], colors = ['xkcd:pumpkin', 'xkcd:slate'],
               linewidths = [2.2, 2.2], custom_ylim=[0, 0.88],
               dpi = 100, show_figure = True,
               save_figure = True, picname = picname, usetex = True)



# Demo: solving the Kuramoto-Sivashinsky equation
# (after Kassam & Trefethen 2005: https://people.maths.ox.ac.uk/trefethen/publication/PDF/2005_111.pdf)
import numpy as np

import joe_lab.joe as joe
import joe_lab.models as joe_models
import joe_lab.initial_states as joe_initial_states

length, T, N, dt = 32.*np.pi, 150., 2**7, 2**-6
stgrid = {'length':length, 'T':T, 'N':N, 'dt':dt}
my_model = joe_models.builtin_model('ks', nonlinear=True)
my_initial_state = joe_initial_states.builtin_initial_state('ks_chaos')
my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=10)

my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=False)

# produce plots and movies
my_sim.hov_plot(cmap='cmo.solar', fieldname='u', show_figure=True, save_figure=True, usetex=True)
#my_sim.save_movie(dpi=200, fps=200, usetex=False, fieldcolor='xkcd:dark orange', fieldname='u')
#my_sim.save_combomovie(dpi=200, usetex=False, fieldcolor='xkcd:dark orange', speccolor='xkcd:dark magenta', fieldname='u')

# report error in first moment
my_sim.get_fm()
error = np.amax(my_sim.fm_error)
print('Maximum error in first moment = %.2E' % error)

# do refinement study to verify accuracy
#nmin, nmax = 1, 9
#Ns = np.array([2**6, 2**7, 2**8])
#dts = np.flip(np.logspace(-nmax, -nmin, num=nmax-nmin+1, base=2.))
#do_refinement_study(my_model, my_initial_state, length, T, Ns, dts, bc='periodic', show_figure=True, save_figure=True, usetex=True)



import numpy as np

from .utils import my_fft, my_ifft

# create all the stuff we need to implement the sponge layer (absorbing layer/segment near bdry where
# artifical damping turns on)

# first, create a function that gives the damping coefficient a la Lu/Trogdon 2023.
# TODO: should this be made to work on both sides of the domain rather than just one?
def damping_coeff_lt(x, sponge_params):
    amp = 1.

    l_endpt = sponge_params['l_endpt']  # -length * 0.5 + 0.5 * length * 0.1

    r_endpt = sponge_params['r_endpt']  # l_endpt + 0.01 * length

    w = sponge_params['width']  # (2 ** -6) * length / 100.

    out = 0.5 * (np.tanh(w * (x - l_endpt)) + 1.) - 0.5 * (np.tanh(w * (x - r_endpt)) + 1.)

    return amp * out


# create a function that gives the damping coefficient a la Bronski 1998.
# TODO: update this! needs to play nicely with sponge params.
def damping_coeff_bronski(x, length, delta=0.1):
    # left endpoint
    lep = -0.5 * length

    # right endpoint
    rep = 0.5 * length

    condlist = [((lep + delta <= x) & (x <= rep - delta)), ((lep <= x) & (x < lep + delta)),
                ((rep - delta < x) & (x <= rep))]

    w = np.pi / (2. * delta)

    funclist = [lambda x: 0, lambda x: 2. * np.cos(w * (x - lep)), lambda x: 2. * np.cos(w * (rep - x))]

    out = np.piecewise(x, condlist, funclist)

    return out


# create the Rayleigh damping term that can be added to the forcing
# syntax is inputs is the same as that for fourier_forcing
def rayleigh_damping(V, x, length, sponge_params, complex=False):

    N = x.size

    if int(V.size - 2) == N or int(0.5*V.size) == N:

        pass

    else:

        raise TypeError("The array V must be 2+(size of the array x) if our soln is real, "
                        "or 2*(size of the array x) if our soln is complex."
                        " Size of V = ", int(V.size), "size of x = ", x.size)

    if complex:

        NN = N

    else:

        NN = int(0.5*N)+1

    V = np.reshape(V, (2*NN))

    v = my_ifft(V[NN:], complex=complex)  # only ifft last NN entries of V because of storage conventions

    beta = damping_coeff_lt(x, sponge_params)+damping_coeff_lt(-x, sponge_params)
    out = 1j * np.zeros(int(2*NN), dtype=float)
    out[NN:] = my_fft(-1. * beta * v, complex=complex)

    return out


# helper function to clip the "spongeless" part of an array
def clip_spongeless(z, sfrac):
    delta = 0.5 * (1. - sfrac)
    N = np.shape(z)[-1]
    out = z[..., int(delta * N):int((1. - delta) * N) + 1]
    return out



import os

from typing import ContextManager, Optional
from alive_progress import alive_bar

#https://stackoverflow.com/questions/44851940/python-cli-progress-bar-spinner-without-iteration
def spinner(title: Optional[str] = None) -> ContextManager:
    """
    Context manager to display a spinner while a long-running process is running.

    Usage:
        with spinner("Fetching data..."):
            fetch_data()

    Args:
        title: The title of the spinner. If None, no title will be displayed.
    """
    return alive_bar(monitor=None, stats=None, title=title, force_tty=True)

import numpy as np
import matplotlib.pyplot as plt
import cmocean.cm as cmo
import matplotlib.animation as animation
os.environ["XDG_SESSION_TYPE"] = "xcb" # prevents a warning from being thrown up for certain linux users, thanks to
# https://stackoverflow.com/questions/69828508/warning-ignoring-xdg-session-type-wayland-on-gnome-use-qt-qpa-platform-wayland
from scipy.interpolate import CubicSpline
from numpy.fft import fft, fftfreq, fftshift

# create Hovmoeller plot of a scalar field u


def hov_plot(x, t, u, fieldname, umin=None, umax=None, dpi=100, show_figure=True, save_figure=False, picname="",
             cmap=cmo.haline, usetex=True):

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False

    if umin is None:
        umin = np.amin(u)
    if umax is None:
        umax = np.amax(u)

    levels = np.linspace(umin, umax, num=300)

    CF = plt.contourf(x, t, u, cmap=cmap, levels=levels)

    # axis labels
    if usetex:
        plt.xlabel(r"$x$", fontsize=22, color='k')
        plt.ylabel(r"$t$", fontsize=22, color='k')

    else:
        plt.xlabel(r"x", fontsize=22, color='k')
        plt.ylabel(r"t", fontsize=22, color='k')

    plt.tick_params(axis='x', which='both', top=False, color='k')
    plt.xticks(fontsize=16, rotation=0, color='k')
    plt.tick_params(axis='y', which='both', right=False, color='k')
    plt.yticks(fontsize=16, rotation=0, color='k')

    plt.xlim([np.amin(x), np.amax(x)])

    # make colorbar
    cbar = plt.colorbar(CF, format='%.2f')
    cbar.ax.tick_params(labelsize=16, color='k')
    plt.clim(umin, umax)

    if usetex:
        cbarlabel_str = r'$u(x,t)$'.replace('u', str(fieldname))

    else:
        cbarlabel_str = 'u(x,t)'.replace('u', str(fieldname))

    cbar.ax.set_ylabel(cbarlabel_str, fontsize=22, color='k')

    # the final piece of the colorbar defn is to change the colorbar ticks to an acceptable color.
    # This is not so easy, and relies on the thread at
    # https://stackoverflow.com/questions/9662995/matplotlib-change-title-and-colorbar-text-and-tick-colors
    cbytick_obj = plt.getp(cbar.ax.axes, 'yticklabels')
    plt.setp(cbytick_obj, color='k')

    plt.tight_layout()

    if save_figure is True:

        # add the folder "joe_visuals" to our path... more on this below
        my_path = os.path.join("joe_visuals")

        # first, if the folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        # and save the fig
        plt.savefig('joe_visuals/' + picname, bbox_inches='tight', dpi=dpi)

    else:

        pass

    if show_figure is True:

        plt.show()

    else:

        pass

    plt.close()

# create a nice 2D plot of y vs. x

def nice_plot(x, y, xlabel, ylabel, dpi=100, custom_ylim=None, show_figure=True, save_figure=False, picname="", linestyle='solid',
              color='xkcd:cerulean', usetex=True):

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False

    fig, ax = plt.subplots()

    plt.plot(x, y, linestyle = linestyle, color=color, linewidth='2')

    plt.xlim([np.amin(x), np.amax(x)])

    if custom_ylim == None:

        pass

    else:

        plt.ylim(custom_ylim)

    # axis labels
    plt.xlabel(xlabel, fontsize=22, color='k')
    plt.ylabel(ylabel, fontsize=22, color='k')

    plt.tick_params(axis='x', which='both', top='off', color='k')
    plt.xticks(fontsize=16, rotation=0, color='k')
    plt.tick_params(axis='y', which='both', right='off', color='k')
    plt.yticks(fontsize=16, rotation=0, color='k')

    plt.tight_layout()

    if save_figure is True:

        # add the folder "joe_visuals" to our path... more on this below
        my_path = os.path.join("joe_visuals")

        # first, if the folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        # and save the fig
        plt.savefig('joe_visuals/' + picname, bbox_inches='tight', dpi=dpi)

    else:

        pass

    if show_figure is True:

        plt.show()

    else:

        pass

    plt.close()

# plot a bunch of stuff on the same axis.
def nice_multiplot(xs, ys, xlabel, ylabel, curvelabels, linestyles, colors, linewidths, custom_ylim = None,
                   dpi=100, show_figure=True,
                   save_figure=False, picname="", usetex=True):

    if len(xs) != len(ys) or len(xs) != len(curvelabels):

        raise ValueError('xs, ys, and curvelabels must be lists of the same length.')

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False

    fig, ax = plt.subplots()

    for m in range(0, len(xs)):

        plt.plot(xs[m], ys[m], linestyle=linestyles[m], color=colors[m], linewidth=linewidths[m],
                 label=curvelabels[m])

    plt.xlim([np.amin(xs), np.amax(xs)])

    if custom_ylim == None:

        pass

    else:

        plt.ylim(custom_ylim)

    ax.legend(fontsize=14, loc='best')

    # axis labels
    plt.xlabel(xlabel, fontsize=22, color='k')
    plt.ylabel(ylabel, fontsize=22, color='k')

    plt.tick_params(axis='x', which='both', top='off', color='k')
    plt.xticks(fontsize=16, rotation=0, color='k')
    plt.tick_params(axis='y', which='both', right='off', color='k')
    plt.yticks(fontsize=16, rotation=0, color='k')

    plt.tight_layout()

    if save_figure is True:

        # add the folder "joe_visuals" to our path... more on this below
        my_path = os.path.join("joe_visuals")

        # first, if the folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        # and save the fig
        plt.savefig('joe_visuals/' + picname, bbox_inches='tight', dpi=dpi)

    else:

        pass

    if show_figure is True:

        plt.show()

    else:

        pass

    plt.close()

def nice_hist(data, xlabel, dpi=100, show_figure=True, save_figure=False, picname="",
              color='xkcd:deep pink', usetex=True):

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False

    fig = plt.figure()
    plt.hist(data, color=color)
    fig.set_size_inches(8, 6)
    plt.tick_params(axis='x', which='both', top=False, color='k')
    plt.xticks(fontsize=20, rotation=0, color='k')
    plt.tick_params(axis='y', which='both', right=False, color='k')
    plt.yticks(fontsize=20, rotation=0, color='k')

    plt.xlabel(xlabel, fontsize=22, color='k')

    plt.ylabel(r"Number of Instances", fontsize=22, color='k')
    plt.xlim([np.amin(data), np.amax(data)])

    plt.tight_layout()

    if save_figure is True:

        # add the folder "joe_visuals" to our path... more on this below
        my_path = os.path.join("joe_visuals")

        # first, if the folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        # and save the fig
        plt.savefig('joe_visuals/' + picname, bbox_inches='tight', dpi=dpi)

    else:

        pass

    if show_figure is True:

        plt.show()

    else:

        pass

    plt.close()

# code for plotting results of refinement study. The input syntax is pretty ugly but it gets the job done and prevents
# the main script from getting too cluttered
def plot_refinement_study(model, initial_state, length, T, Ns, dts, errors, method_kw='etdrk4', bc='periodic',
                        show_figure=True, save_figure=False, usetex=True, dpi=400):

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False

    fig, ax = plt.subplots()

    num_Ns = np.size(Ns)

    dts = 0.5 * dts

    # define the cycler
    my_cycler = (
            plt.cycler(color=['xkcd:slate', 'xkcd:raspberry', 'xkcd:goldenrod', 'xkcd:deep green'])
            + plt.cycler(lw=[3.5, 3, 2.5, 2])
            + plt.cycler(linestyle=['dotted', 'dashed', 'solid', 'dashdot'])
            + plt.cycler(marker=['v', '*', 'o', 'P'])
            + plt.cycler(markersize=[8, 12, 8, 8])
    )

    ax.set_prop_cycle(my_cycler)

    for m in range(0, num_Ns):
        if usetex:
            plt.loglog(dts, errors[m, :], label=r'$N = z$'.replace('z', str(Ns[m])))
        # ^ an awesome trick from
        # https://stackoverflow.com/questions/33786332/matplotlib-using-variables-in-latex-expressions
        # was used to get the labels working as above
        else:
            plt.loglog(dts, errors[m, :], label='N = z'.replace('z', str(Ns[m])))

    ax.legend(fontsize=16)

    if usetex:
        plt.xlabel(r"$\Delta t$", fontsize=26, color='k')
        plt.ylabel(r"Absolute Error", fontsize=26, color='k')
    else:
        plt.xlabel("Δt", fontsize=26, color='k')
        plt.ylabel("Absolute Error", fontsize=26, color='k')

    plt.tick_params(axis='x', which='both', top='off', color='k')
    plt.xticks(fontsize=16, rotation=0, color='k')
    plt.tick_params(axis='y', which='both', right='off', color='k')
    plt.yticks(fontsize=16, rotation=0, color='k')

    plt.tight_layout()

    if save_figure is True:

        # add the folder "joe_visuals" to our path
        my_path = os.path.join("joe_visuals")

        # first, if the folder doesn't exist, make it
        if not os.path.isdir(my_path):
            os.makedirs(my_path)

        # and now we can save the fig
        if bc == 'sponge_layer':
            absorbing_layer = True
        elif bc == 'periodic':
            absorbing_layer = False

        my_string = ('_length=%.1f_T=%.1f' % (
            length, T) + '_modelkw=' + model.model_kw + '_ICkw=' + initial_state.initial_state_kw + '_method_kw='
                     + method_kw + '_nonlinear=' + str(model.nonlinear) + '_abslayer=' + str(absorbing_layer))

        picname = 'refinement_study' + my_string + '.png'
        plt.savefig('joe_visuals/' + picname, bbox_inches='tight', dpi=dpi)

    else:

        pass

    if show_figure is True:

        plt.show()

    else:

        pass

    plt.close()

# create a movie from a scalar field u(t,x) sampled at various times.

def save_movie(u, x, length, dt, fieldname, ndump, filename, fps=200, periodic=True, usetex=True,
               fieldcolor='xkcd:ocean green', dpi=100):
    # Create movie file in mp4 format. Warning: this is very slow!

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError: # catch a user error thinking they have tex when they don't
        usetex = False

    fig = plt.figure()

    umin = np.amin(u)
    umax = np.amax(u)

    if umin < 0:

        umin *= 1.05

    else:

        umin *=0.95

    if umax > 0:

        umax *= 1.05

    else:

        umax *= 0.95

    #from scipy.integrate import simps

    # coeff = simps(np.abs(u[0, :]),x)  #TODO: multiplying by a factor of 4 gives a solid result , but for BBM 6 is # the
    # correct "back of the envelope" value.

    ax = plt.axes(xlim=(np.amin(x), np.amax(x)), ylim=(umin, umax))
    #ax = plt.axes(xlim=(-60, 25), ylim=(umin, umax))

    fig.set_size_inches(8, 6)

    # use cubic spline interpolation to smooth the data

    # note that spline interpolation is necessary to combat the Runge phenomenon since we have a uniform grid! Naive
    # use of barycentric interpolation is not an option.

    if periodic:

        # add endpoint
        x_end = np.append(x, 0.5 * length)

        [M, N] = np.shape(u)

        u_end = np.zeros([M, N + 1], dtype=float)

        u_end[:, 0:N] = u

        u_end[:, -1] = np.copy(u[:, 0])  # enforce periodicity

        # create the interpolating spline polynomial using scipy interpolation package

        poly = CubicSpline(x_end, u_end, axis=1, bc_type='periodic')

    else:

        poly = CubicSpline(x, u, axis=1)

    # now we can actually do the upsampling

    NN = 2000  # number of points to evaluate interpolant at

    xx = np.linspace(np.amin(x), np.amax(x), NN, endpoint=True)

    uu = poly(xx)

    x = xx

    u = uu

    ax.grid('True')

    line, = ax.plot([], [], linewidth=2, color=fieldcolor, label=fieldname)
    #bound_line_up, = ax.plot([], [], linewidth=2, color='xkcd:emerald', linestyle='dashed',
                            # label='$\|u_0\|_{L^{1}_{x}} (1+t)^{-1/3}$')
    #bound_line_down, = ax.plot([], [], linewidth=2, color='xkcd:emerald', linestyle='dashed')

    timer = fig.canvas.new_timer(interval=100)
    timer.add_callback(u, ax)
    timer.start()

    #ax.legend(fontsize=20, loc='upper left')

    plt.tick_params(axis='x', which='both', top=False, color='k')
    plt.xticks(fontsize=20, rotation=0, color='k')
    plt.tick_params(axis='y', which='both', right=False, color='k')
    plt.yticks(fontsize=20, rotation=0, color='k')

    if usetex:
        plt.xlabel(r"$x$", fontsize=22, color='k')
    else:
        plt.xlabel(r"x", fontsize=22, color='k')

    # Animation function.
    def animate(i):
        line.set_data(x, u[i, :])

        tplot = i * dt * ndump

        #bound_line_up.set_data(x, coeff / ((1 + tplot) ** (1. / 3.)))
        #bound_line_down.set_data(x, -coeff / ((1 + tplot) ** (1. / 3.)))

        #plt.title('$t=%.2f$' % tplot, fontsize=22)

        if usetex:
            ylabel_str = r'$u(x,t=%.2f)$'.replace('u', str(fieldname)) % tplot

        else:
            ylabel_str = 'u(x,t=%.2f)'.replace('u', str(fieldname)) % tplot

        ax.set_ylabel(ylabel_str, fontsize=22)

        plt.xlim([np.amin(x), np.amax(x)])
        #plt.xlim = ([-60., 25.])

        """
        if i%400==0 or i == 800 or i==1000:

            plt.savefig('joe_visuals/' + 'frame=%.1f' %i + '.png', bbox_inches='tight', dpi=600)

        else:

            pass 
        """

        plt.tight_layout()

        return line,

    anim = animation.FuncAnimation(fig, animate, np.shape(u)[0], blit=False)

    #plt.savefig('joe_visuals/' + 'frame'+ '.png', bbox_inches='tight', dpi=100)

    # add the folder "joe_visuals" to our path... more on this below
    my_path = os.path.join("joe_visuals")

    # first, if the folder doesn't exist, make it
    if not os.path.isdir(my_path):
        os.makedirs(my_path)

    anim.save('joe_visuals/' + filename, writer='ffmpeg', fps=fps, extra_args=['-vcodec', 'libx264'], dpi=dpi)

    plt.close()

def save_combomovie(u, x, length, dt, ndump, filename, fieldname,
                    fps=200, periodic=True, usetex=True, fieldcolor='xkcd:ocean green', speccolor='xkcd: dark magenta', dpi=100):
    # Create movie file in mp4 format. Warning: this is very slow!

    plt.rcParams["font.family"] = "serif"

    try:
        plt.rc('text', usetex=usetex)

    except RuntimeError:  # catch a user error thinking they have tex when they don't
        usetex = False


    fig = plt.figure()

    umin = 1.05 * np.amin(u)
    umax = 1.05 * np.amax(u)

    ax = plt.axes(xlim=(np.amin(x), np.amax(x)), ylim=(umin, umax))

    # create insert axes
    v = np.absolute(fft(u, axis=1)) ** 2

    N = np.shape(x)[0]

    k = fftshift(2. * np.pi * N * fftfreq(N) / length)

    kmin = np.amin(k)
    kmax = np.amax(k)

    vmin = 1.05 * np.amin(v)
    vmax = 1.05 * np.amax(v)

    v = fftshift(v, axes=1)

    ins = ax.inset_axes([0.15, 0.64, 0.3, 0.3], xlim=(kmin, kmax), ylim=(vmin, vmax))
    # phi4 vals: [0.69, 0.685, 0.3, 0.3]

    # use cubic spline interpolation to smooth the state data

    # note that spline interpolation is necessary to combat the Runge phenomenon since we have a uniform grid! Naive
    # use of barycentric interpolation is not an option.

    if periodic:

        # add endpoint
        x_end = np.append(x, 0.5 * length)

        [M, N] = np.shape(u)

        u_end = np.zeros([M, N + 1], dtype=float)

        u_end[:, 0:N] = u

        u_end[:, -1] = np.copy(u[:, 0])  # enforce periodicity

        # create the interpolating spline polynomial using scipy interpolation package

        poly = CubicSpline(x_end, u_end, axis=1, bc_type='periodic')

    else:

        poly = CubicSpline(x, u, axis=1)

    # now we can actually do the spatial upsampling

    NN = 2000  # number of points to evaluate interpolant at

    xx = np.linspace(np.amin(x), np.amax(x), NN, endpoint=True)

    uu = poly(xx)

    x = xx

    u = uu

    # now make smaller insert graph plotting data

    poly = CubicSpline(k, v, axis=1)  # ignore periodic flag here, it's not really worth the extra effort

    NN = 3600  # number of points to evaluate interpolant at

    kk = np.linspace(kmin, kmax, NN, endpoint=True)

    vv = poly(kk)

    k = kk

    v = vv

    ax.grid(True)
    if usetex:
        ax.set_xlabel('$x$', fontsize=22)
    else:
        ax.set_xlabel('x', fontsize=22)

    ins.grid(False)
    if usetex:
        ins.set_xlabel('$k$', fontsize=11)
    else:
        ins.set_xlabel('k', fontsize=11)

    line, = ax.plot([], [], linewidth=2, color=fieldcolor)
    iline, = ins.plot([], [], linewidth=1., color=speccolor)

    timer = fig.canvas.new_timer(interval=100)
    timer.add_callback(u, ax)
    timer.start()

    # Animation function.
    def animate(i):
        line.set_data(x, u[i, :])
        iline.set_data(k, v[i, :])

        tplot = i * dt * ndump

        if usetex:
            ylabel_str = r'$u(x,t=%.2f)$'.replace('u', str(fieldname)) % tplot

        else:
            ylabel_str = 'u(x,t=%.2f)'.replace('u', str(fieldname)) % tplot

        ax.set_ylabel(ylabel_str, fontsize=22)

        if usetex:
            in_ylabel_str = r'$|\widehat{u}|^2(k,t=%.2f)$'.replace('u', str(fieldname)) % tplot

        else:
            in_ylabel_str = 'Pow[u](k,t=%.2f)'.replace('u', str(fieldname)) % tplot

        ins.set_ylabel(in_ylabel_str, fontsize=11)

        plt.tight_layout()

        return line,

    anim = animation.FuncAnimation(fig, animate, np.shape(u)[0], blit=False)

    # add the folder "joe_visuals" to our path... more on this below
    my_path = os.path.join("joe_visuals")

    # first, if the folder doesn't exist, make it
    if not os.path.isdir(my_path):
        os.makedirs(my_path)

    anim.save('joe_visuals/' + filename, fps=fps, extra_args=['-vcodec', 'libx264'], dpi=dpi)

    plt.close()



import joe_lab.joe as joe
import joe_lab.models as joe_models
import joe_lab.initial_states as joe_initial_states

length, T, N, dt = 100.,50., 2**9, 1e-2

stgrid = {'length': length, 'T': T, 'N': N, 'dt': dt}
my_model = joe_models.builtin_model('sinegordon', nonlinear=True)
my_initial_state = joe_initial_states.builtin_initial_state('sinegordon_soliton_interaction') #sinegordon_soliton_interaction

my_sim = joe.simulation(stgrid, my_model, my_initial_state, bc='periodic', ndump=20)

#my_sim.plot_initial_condition(color='xkcd:deep magenta', usetex=True, show_figure=True, save_figure=True)

my_sim.load_or_run(method_kw='etdrk4', print_runtime=True, save=False)

# produce plots and movies
my_sim.hov_plot(cmap='cmo.dense', fieldname='u', show_figure=True, save_figure=True, usetex=True)
#my_sim.save_movie(dpi=200, fps=90, usetex=False, fieldcolor='xkcd:deep magenta', fieldname='u')

"""
import numpy as np

nmin, nmax = 4, 9
Ns = np.array([2**7, 2**8, 2**9, 2**10])
dts = np.flip(np.logspace(-nmax, -nmin, num=nmax-nmin+1, base=2.))

joe.do_refinement_study(my_model, my_initial_state, length, T, Ns, dts, bc='periodic', method_kw='etdrk4',
                    show_figure=True, save_figure=True, usetex=True, fit_min=0, fit_max=4)
"""
